"""gustaf/gustaf/io/mfem.py.

io functions for mfem. Supports simple linear elements (straight meshes)
For detailed information, see: https://mfem.org/mesh-
format-v1.0/#straight-meshes
"""

import numpy as np

from gustaf import settings
from gustaf.faces import Faces
from gustaf.volumes import Volumes

geometry_types = {
        "POINT": 0,
        "SEGMENT": 1,
        "TRIANGLE": 2,
        "SQUARE": 3,
        "TETRAHEDRON": 4,
        "CUBE": 5,
        "PRISM": 6
}


def load(fname):
    """Load mesh in MFEM format. Loads vertices and their connectivity.
    Currently cannot process boundary.

    Parameters
    ------------
    fname: str

    Returns
    ------------
    mesh
    """

    def extract_values(fname, start_index, n_lines, total_lines, dtype):
        """Extract information from file. Reads [n_lines] lines from file
        [fname] starting at line [start_index] and returns array of type
        [dtype].

        Parameters
        ------------
        fname: str
        start_index: int
        n_lines: int
        total_lines: int
            Number of total lines of file
        dtype: (NumPy) data type

        Returns
        ------------
        NumPy Array
        """
        end_index = total_lines - (start_index + n_lines + 2)
        return np.genfromtxt(
                fname,
                delimiter=" ",
                skip_header=start_index,
                skip_footer=end_index,
                dtype=dtype
        )

    with open(fname, "r") as f:
        lines = f.readlines()
        total_lines = len(lines)

        # Read values of keywords
        keywords = ["dimension", "elements", "boundary", "vertices"]
        indices = [lines.index(f"{keyword}\n") for keyword in keywords]
        dimension, n_elements, n_boundaries, n_vertices = [
                int(lines[index + 1]) for index in indices
        ]
        vdim = int(lines[indices[-1] + 2])
        # Extract values
        elements = extract_values(
                fname, indices[1] + 2, n_elements, total_lines,
                settings.INT_DTYPE
        )
        if elements.shape[0] != n_elements:
            raise ValueError("Number of elements do not match.")
        boundary = extract_values(
                fname, indices[2] + 2, n_boundaries, total_lines,
                settings.INT_DTYPE
        )
        if boundary.shape[0] != n_boundaries:
            raise ValueError("Number of boundaries do not match.")
        vertices = extract_values(
                fname, indices[3] + 3, n_vertices, total_lines,
                settings.FLOAT_DTYPE
        )
        if vertices.shape != (n_vertices, vdim):
            raise ValueError("Number of vertices do not match.")

        connectivity = elements[:, 2:]
        if dimension == 2:
            mesh = Faces(vertices=vertices, faces=connectivity)
        elif dimension == 3:
            mesh = Volumes(vertices=vertices, volumes=connectivity)

        return mesh


def export(mesh, fname):
    """Export mesh in MFEM format. Supports 2D triangle and quadrilateral
    meshes. Does not support different element attributes or difference in
    vertex dimension and mesh dimension.

    Parameters
    ------------
    mesh: Faces
    fname: str

    Returns
    ------------
    None
    """
    # Basic infos
    nvertices, dim = mesh.vertices.shape

    def format_array(array):
        """Format NumPy array into string. Each entry in a row is separated by
        a blank space and every row is separated by a new line.

        Parameters
        ------------
        array: NumPy array

        Returns
        ------------
        str
        """
        row_strings = [" ".join(map(str, row)) for row in array]
        return "\n".join(row_strings)

    # Export 2D mesh
    if dim == 2:
        # Elements
        element_attribute = 1  # Other numbers not yet supported
        elements = mesh.elements
        n_elements, n_element_vertices = elements.shape
        if n_element_vertices == 3:
            geometry_type = geometry_types["TRIANGLE"]
        elif n_element_vertices == 4:
            geometry_type = geometry_types["SQUARE"]
        else:
            raise NotImplementedError(
                    "Sorry, we cannot export mesh with elements "
                    f"with {n_element_vertices} vertices."
            )
        e = np.ones((n_elements, 1), dtype=settings.INT_DTYPE)
        elements_array = np.hstack(
                (element_attribute * e, geometry_type * e, elements)
        )
        elements_array_string = format_array(elements_array)
        elements_string = f"elements\n{n_elements}\n"
        elements_string += f"{elements_array_string}\n\n"

        # Boundary
        edges = mesh.edges()
        nboundary_edges = sum(map(len, mesh.BC.values()))
        boundary_array = np.empty(
                (nboundary_edges, 4), dtype=settings.INT_DTYPE
        )
        startrow = 0
        # Add boundary one by one as SEGMENTs
        for bid, edgeids in mesh.BC.items():
            nedges = len(edgeids)
            e = np.ones(nedges).reshape(-1, 1)
            vertex_list = edges[edgeids, :]
            boundary_array[startrow:(startrow + nedges), :] = np.hstack(
                    (int(bid) * e, geometry_types["SEGMENT"] * e, vertex_list)
            )
            startrow += nedges

        boundary_array_string = format_array(boundary_array)
        boundary_string = f"boundary\n{nboundary_edges}\n"
        boundary_string += f"{boundary_array_string}\n\n"

        # Vertices
        vdim = 2  # Currently only option
        vertices_array_string = format_array(mesh.vertices)
        vertices_string = f"vertices\n{nvertices}\n{vdim}\n"
        vertices_string += f"{vertices_array_string}"

        with open(fname, "w") as f:
            f.write("MFEM mesh v1.0\n\n")
            f.write(f"dimension\n{dim}\n\n")
            f.write(elements_string)
            f.write(boundary_string)
            f.write(vertices_string)

    # Export 3D mesh
    else:
        raise NotImplementedError(
                f"Sorry, we cannot export mesh of dimension {dim}."
        )



"""gustaf/gustaf/volumes.py."""

import numpy as np

from gustaf import utils
from gustaf import helpers
from gustaf import settings
from gustaf.faces import Faces


class Volumes(Faces):

    kind = "volume"

    const_faces = helpers.raise_if.invalid_inherited_attr(
            Faces.const_faces,
            __qualname__,
            property_=True,
    )
    update_faces = helpers.raise_if.invalid_inherited_attr(
            Faces.update_edges,
            __qualname__,
            property_=False,
    )

    __slots__ = (
            "_volumes",
            "_const_volumes",
    )

    def __init__(
            self,
            vertices=None,
            volumes=None,
            elements=None,
    ):
        """Volumes. It has vertices and volumes. Volumes could be tetrahedrons
        or hexahedrons.

        Parameters
        -----------
        vertices: (n, d) np.ndarray
        volumes: (n, 4) or (n, 8) np.ndarray
        """
        super().__init__(vertices=vertices)
        if volumes is not None:
            self.volumes = volumes
        elif elements is not None:
            self.volumes = elements

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def faces(self):
        """Faces here aren't main property. So this needs to be computed.

        Parameters
        -----------
        None

        Returns
        --------
        faces: (n, 3) or (n, 4) np.ndarray
        """
        whatami = self.whatami
        faces = None
        if whatami.startswith("tet"):
            faces = utils.connec.tet_to_tri(self.volumes)
        elif whatami.startswith("hexa"):
            faces = utils.connec.hexa_to_quad(self.volumes)

        return faces

    @classmethod
    def whatareyou(cls, volume_obj):
        """overwrites Faces.whatareyou to tell you is this volume is tet or
        hexa.

        Parameters
        -----------
        volume_obj: Volumes

        Returns
        --------
        whatareyou: str
        """
        if not cls.kind.startswith(volume_obj.kind):
            raise TypeError("Given obj is not {cls.__qualname__}")

        if volume_obj.volumes.shape[1] == 4:
            return "tet"

        elif volume_obj.volumes.shape[1] == 8:
            return "hexa"

        else:
            raise ValueError(
                    "Invalid volumes connectivity shape. It should be (n, 4) "
                    f"or (n, 8), but given: {volume_obj.volumes.shape}"
            )

    @property
    def volumes(self):
        """Returns volumes.

        Parameters
        -----------
        None

        Returns
        --------
        volumes: (n, 4) or (n, 8) np.ndarray
        """
        return self._volumes

    @volumes.setter
    def volumes(self, vols):
        """volumes setter. Similar to vertices, this will be a tracked array.

        Parameters
        -----------
        vols: (n, 4) or (n, 8) np.ndarray

        Returns
        --------
        None
        """
        if vols is not None:
            utils.arr.is_one_of_shapes(
                    vols,
                    ((-1, 4), (-1, 8)),
                    strict=True,
            )

        self._volumes = helpers.data.make_tracked_array(
                vols,
                settings.INT_DTYPE,
        )
        # same, but non-writeable view of tracked array
        self._const_volumes = self._volumes.view()
        self._const_volumes.flags.writeable = False

    @property
    def const_volumes(self):
        """Returns non-writeable view of volumes.

        Parameters
        -----------
        None

        Returns
        --------
        const_volumes: (n, 4) or (n, 8) np.ndarray
        """
        return self._const_volumes

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def sorted_volumes(self):
        """Sort volumes along axis=1.

        Parameters
        -----------
        None

        Returns
        --------
        volumes_sorted: (volumes.shape) np.ndarray
        """
        volumes = self._get_attr("volumes")

        return np.sort(volumes, axis=1)

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def unique_volumes(self):
        """Returns a namedtuple of unique volumes info. Similar to
        unique_edges.

        Parameters
        -----------
        None

        Returns
        --------
        unique_info: Unique2DIntegers
          valid attribut4es are {values, ids, inverse, counts}
        """
        unique_info = utils.connec.sorted_unique(
                self.sorted_volumes(),
                sorted_=True,
        )

        volumes = self._get_attr("volumes")

        unique_info.values[:] = volumes[unique_info.ids]

        return unique_info

    def update_volumes(self, *args, **kwargs):
        """Alias to update_elements."""
        self.update_elements(*args, **kwargs)

    def tofaces(self, unique=True):
        """Returns Faces obj.

        Parameters
        -----------
        unique: bool
          Default is True. If True, only takes unique faces.

        Returns
        --------
        faces: Faces
        """
        return Faces(
                self.vertices,
                faces=self.unique_faces().values if unique else self.faces()
        )



"""gustaf/create/volumes.py
Routines to create volumes.
"""

import numpy as np

from gustaf.volumes import Volumes
from gustaf import utils
from gustaf import create


def box(bounds=[[0., 0., 0.], [1., 1., 1.]], resolutions=[2, 2, 2]):
    """Create structured hexahedron block mesh.

    Parameters
    -----------
    bounds: (2, 3) array
        Minimum and maximum coordinates.
    resolutions: (3) array
        Vertex count in each dimension.

    Returns
    --------
    volume_mesh: Volumes
    """

    if np.array(bounds).shape != (2, 3):
        raise ValueError("Bounds must have a dimension of (2, 3).")
    if len(resolutions) != 3:
        raise ValueError("Resolutions must have three entries.")
    if not np.greater(resolutions, 1).all():
        raise ValueError("All resolution values must be at least 2.")

    vertex_mesh = create.vertices.raster(bounds, resolutions)
    connectivity = utils.connec.make_hexa_volumes(resolutions)
    volume_mesh = Volumes(vertex_mesh.vertices, connectivity)

    return volume_mesh



"""gustaf/gustaf/faces.py."""
import numpy as np

from gustaf import settings
from gustaf import utils
from gustaf.edges import Edges
from gustaf import helpers


class Faces(Edges):

    kind = "face"

    const_edges = helpers.raise_if.invalid_inherited_attr(
            Edges.const_edges,
            __qualname__,
            property_=True,
    )
    update_edges = helpers.raise_if.invalid_inherited_attr(
            Edges.update_edges,
            __qualname__,
            property_=False,
    )
    dashed = helpers.raise_if.invalid_inherited_attr(
            Edges.const_edges,
            __qualname__,
            property_=False,
    )

    __slots__ = (
            "_faces",
            "_const_faces",
            "BC",
    )

    def __init__(
            self,
            vertices=None,
            faces=None,
            elements=None,
    ):
        """Faces. It has vertices and faces. Faces could be triangles or
        quadrilaterals.

        Parameters
        -----------
        vertices: (n, d) np.ndarray
        faces: (n, 3) or (n, 4) np.ndarray
        """
        super().__init__(vertices=vertices)
        if faces is not None:
            self.faces = faces

        elif elements is not None:
            self.faces = elements

        self.BC = dict()

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def edges(self):
        """Edges from here aren't main property. So this needs to be computed.

        Parameters
        -----------
        None

        Returns
        --------
        edges: (n, 2) np.ndarray
        """
        self._logd("computing edges")
        faces = self._get_attr("faces")

        return utils.connec.faces_to_edges(faces)

    @property
    def whatami(self, ):
        """Determines whatami.

        Parameters
        -----------
        None

        Returns
        --------
        None
        """
        return type(self).whatareyou(self)

    @classmethod
    def whatareyou(cls, face_obj):
        """classmethod that tells you if the Faces is tri or quad or invalid
        kind.

        Parameters
        -----------
        face_obj: Faces

        Returns
        --------
        whatareyou: str
        """
        if not cls.kind.startswith(face_obj.kind):
            raise TypeError("Given obj is not {cls.__qualname__}")

        if face_obj.faces.shape[1] == 3:
            return "tri"

        elif face_obj.faces.shape[1] == 4:
            return "quad"

        else:
            raise ValueError(
                    "Invalid faces connectivity shape. It should be (n, 3) or "
                    f"(n, 4), but given: {face_obj.faces.shape}"
            )

    @property
    def faces(self, ):
        """Returns faces.

        Parameters
        -----------
        None

        Returns
        --------
        faces
        """
        self._logd("returning faces")
        return self._faces

    @faces.setter
    def faces(self, fs):
        """Faces setter. Similar to veritces, this will be a tracked array.

        Parameters
        -----------
        fs: (n, 2) np.ndarray

        Returns
        --------
        None
        """
        self._logd("setting faces")

        # shape check
        if fs is not None:
            utils.arr.is_one_of_shapes(
                    fs,
                    ((-1, 3), (-1, 4)),
                    strict=True,
            )

        self._faces = helpers.data.make_tracked_array(
                fs,
                settings.INT_DTYPE,
        )
        # same, but non-writeable view of tracekd array
        self._const_faces = self._faces.view()
        self._const_faces.flags.writeable = False

    @property
    def const_faces(self):
        """Returns non-writeable view of faces.

        Parameters
        -----------
        None

        Returns
        --------
        const_faces: (n, 2
        """
        return self._const_faces

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def sorted_faces(self):
        """Similar to edges_sorted but for faces.

        Parameters
        -----------
        None

        Returns
        --------
        sorted_faces: (self.faces.shape) np.ndarray
        """
        faces = self._get_attr("faces")

        return np.sort(faces, axis=1)

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def unique_faces(self):
        """Returns a namedtuple of unique faces info. Similar to unique_edges.

        Parameters
        -----------
        None

        Returns
        --------
        unique_info: Unique2DIntegers
          valid attributes are {values, ids, inverse, counts}
        """
        unique_info = utils.connec.sorted_unique(
                self.sorted_faces(), sorted_=True
        )

        faces = self._get_attr("faces")

        unique_info.values[:] = faces[unique_info.ids]

        return unique_info

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def single_faces(self):
        """Returns indices of very unique faces: faces that appear only once.
        For well constructed volumes, this can be considered as surfaces.

        Parameters
        -----------
        None

        Returns
        --------
        single_faces: (m,) np.ndarray
        """
        unique_info = self.unique_faces()

        return unique_info.ids[unique_info.counts == 1]

    def update_faces(self, *args, **kwargs):
        """Alias to update_elements."""
        self.update_elements(*args, **kwargs)

    def toedges(self, unique=True):
        """Returns Edges obj.

        Parameters
        -----------
        unique: bool
          Default is True. If True, only takes unique edges.

        Returns
        --------
        edges: Edges
        """
        return Edges(
                self.vertices,
                edges=self.unique_edges().values if unique else self.edges()
        )



"""gustaf/create/spline.py.

Frequently used spline shapes generation.
"""

import numpy as np

from gustaf.create.vertices import raster
from gustaf.spline.base import BSpline


def with_bounds(
        parametric_bounds,
        physical_bounds,
        degrees=None,
        num_unique_knots=None,
        nurbs=False,
):
    """Creates spline with given parametric bounds, physical bounds, degrees,
    num_unique_knots. Physical bounds can have less or equal number of
    dimension as parametric bounds. (Greater is not supported)

    Parameters
    -----------
    parametric_bounds: (2, n) array-like
    physical_bounds: (2, n) array-like
    degrees: (n,) array-like
      (Optional) Default is 1 for each dimension.
    num_unique_knots: (n,) array-like
      (Optional) Default is 2 for each dimension.
    nurbs: bool
      (Optional) Default is False.

    Returns
    --------
    spline: BSpline or NURBS
      If `spline` is not availabe, will return dict of corresponding
    """
    # First, prepare for degree 1 spline.
    # KV
    l_bound, u_bound = parametric_bounds
    assert len(l_bound) == len(u_bound),\
        "Length of lower and upper parametric_bounds aren't identical"
    kvs = [[lb, lb, u, u] for lb, u in zip(l_bound, u_bound)]

    # CP
    pl_bound, pu_bound = physical_bounds
    assert len(pl_bound) == len(pu_bound),\
        "Length of upper and lower physical_bounds aren't identical"
    dim_diff = len(l_bound) - len(pl_bound)
    if dim_diff < 0:
        raise ValueError(
                "Sorry, we don't support spline generation with phys_dim > "
                "para_dim."
        )

    cps = raster(physical_bounds, [2] * len(pl_bound)).vertices
    cps = np.repeat(cps, dim_diff + 1, axis=0)

    # Now, make spline
    spl = BSpline(
            knot_vectors=kvs,
            control_points=cps,
            degrees=[1] * len(l_bound),
    )

    # Return early if there's nothing left to do
    if degrees is None and num_unique_knots is None:
        return spl

    # Manipulate to satisfy degrees and num_unique_knots
    # Degrees
    for i, d in enumerate(degrees):
        diff = int(d - 1)
        # continue if desired degree is 1
        if diff == 0:
            continue
        # Elevate degree
        for _ in range(diff):
            spl.elevate_degree(i)

    # Return is there's nothing to do
    if num_unique_knots is None:
        return spl if not nurbs else spl.nurbs

    # Knots
    assert len(num_unique_knots) == len(l_bound),\
        "Length of num_unique_knots and parametric_bounds[0] does not match."
    for i, (nuk, lb, ub) in enumerate(zip(num_unique_knots, l_bound, u_bound)):
        if nuk < .5:
            continue
        new_knots = np.linspace(lb, ub, nuk)[1:-1]
        spl.insert_knots(i, new_knots)

    return spl if not nurbs else spl.nurbs


def with_parametric_bounds(
        parametric_bounds,
        degrees=None,
        num_unique_knots=None,
        nurbs=False,
):
    """Similar to with_bounds. Creates spline that has same parametric and
    physical bounds.

    Parameters
    -----------
    parametric_bounds: (2, n) array-like
    degrees: (n,) array-like
      (Optional) Default is 1 for each dimension.
    num_unique_knots: (n,) array-like
      (Optional) Default is 2 for each dimension.
    nurbs: bool
      (Optional) Default is False.

    Returns
    --------
    spline: BSpline or NURBS
      If `spline` is not availabe, will return dict of corresponding
    """
    return with_bounds(
            parametric_bounds=parametric_bounds,
            physical_bounds=parametric_bounds,
            degrees=degrees,
            num_unique_knots=num_unique_knots,
            nurbs=nurbs,
    )


def with_physical_bounds(
        physical_bounds,
        degrees=None,
        num_unique_knots=None,
        nurbs=None,
):
    """Similar to with_bounds. Creates spline that has same [0, 1]^3 parametric
    bounds.

    Parameters
    -----------
    physical_bounds: (2, n) array-like
    degrees: (n,) array-like
      (Optional) Default is 1 for each dimension.
    num_unique_knots: (n,) array-like
      (Optional) Default is 2 for each dimension.
    nurbs: bool
      (Optional) Default is False.

    Returns
    --------
    spline: BSpline or NURBS
      If `spline` is not availabe, will return dict of corresponding
    """
    dim = len(physical_bounds[0])

    return with_bounds(
            parametric_bounds=[
                    [0. for _ in range(dim)],
                    [1. for _ in range(dim)],
            ],
            physical_bounds=physical_bounds,
            num_unique_knots=num_unique_knots,
            nurbs=nurbs,
    )


def parametric_view(spline):
    """Create parametric view of given spline. Previously called
    `naive_spline()`. Degrees are always 1 and knot multiplicity is not
    preserved. Returns BSpline, as BSpline and NURBS should look the same.

    Parameters
    -----------
    spline: BSpline or NURBS

    Returns
    --------
    para_spline: BSpline
    """
    para_spline = with_parametric_bounds(
            parametric_bounds=spline.knot_vector_bounds,
            degrees=[1] * spline.para_dim,
            num_unique_knots=None,
            nurbs=False,
    )

    # loop through knot vectors and insert missing knots
    for i, kv in enumerate(spline.unique_knots):
        if len(kv) > 2:
            para_spline.insert_knots(i, kv[1:-1])

    return para_spline



"""gustaf/gustaf/show.py.

Everything related to show/visualization.
"""
import numpy as np

from gustaf import settings
from gustaf import utils
from gustaf._base import GustafBase

# @linux it raises error if vedo is imported inside the function.
try:
    import vedo
except ImportError as err:
    # overwrites the vedo module with an object which will throw an error
    # as soon as it is used the first time. This means that any non vedo
    # functionality works as before, but as soon as vedo is used a
    # comprehensive exception will be raised which is understandable in
    # contrast to the possible errors previously possible
    from gustaf.helpers.raise_if import ModuleImportRaiser
    vedo = ModuleImportRaiser("vedo", err)


def show(*gusobj, **kwargs):
    """Shows using appropriate backend.

    Parameters
    -----------
    *gusobj: gustaf objects

    Returns
    --------
    None
    """
    showables = [make_showable(g, **kwargs) for g in gusobj]
    vis_b = settings.VISUALIZATION_BACKEND

    if vis_b.startswith("vedo"):
        return show_vedo(showables, **kwargs)
    elif vis_b.startswith("trimesh"):
        pass
    elif vis_b.startswith("matplotlib"):
        pass
    else:
        raise NotImplementedError


def show_vedo(
        *args,
        **kwargs,
):
    """`vedo.show` wrapper. Each args represent one section of window. In other
    words len(args) == N, where N corresponds to the parameter for vedo.show().

    Parameters
    -----------
    *args: Union[List[Union[gustaf_obj, vedo_obj]], Dict[str, Any]]]
    """
    # vedo plotter parameter
    N = len(args)
    offs = kwargs.get("offscreen", False)
    interac = kwargs.get("interactive", True)
    plt = kwargs.get("vedoplot", None)
    skip_clear = kwargs.get("skip_clear", False)
    close = kwargs.get("close", None)
    size = kwargs.get("size", "auto")
    cam = kwargs.get("cam", None)
    title = kwargs.get("title", "gustaf")
    return_show_list = kwargs.get("return_showable_list", False)

    def clear_vedoplotter(plotter, numrenderers, skipcl=skip_clear):
        """enough said."""
        # for whatever reason it is desired
        if skipcl:
            return None

        for i in range(numrenderers):
            plotter.clear(at=i)

        return None

    def cam_tuple_to_list(dictcam):
        """if entity is tuple, turns it into list."""
        if dictcam is None:
            return None

        for key, value in dictcam.items():
            if isinstance(value, tuple):
                dictcam[key] = list(value)

        return dictcam

    # get plotter
    if plt is None:
        plt = vedo.Plotter(
                N=N, sharecam=False, offscreen=offs, size=size, title=title
        )

    else:
        # check if plt has enough Ns
        trueN = np.prod(plt.shape)
        clear_vedoplotter(plt, trueN)  # always clear.
        if trueN != N:
            utils.log.warning(
                    "Number of args exceed given vedo.Plotter's capacity.",
                    "Assigning a new one",
            )
            title = plt.title
            if close:  # only if it is explicitly stated
                plt.close()  # Hope that this truely releases..
            # assign a new one
            plt = vedo.Plotter(
                    N=N,
                    sharecam=False,
                    offscreen=offs,
                    size=size,
                    title=title
            )

    # loop and plot
    for i, arg in enumerate(args):
        # form valid input type.
        if isinstance(arg, dict):
            showlist = list(arg.values())
        elif isinstance(arg, list):
            showlist = arg.copy()
        else:
            # raise TypeError(
            #     "For vedo_show, only list or dict is valid input")
            utils.log.debug(
                    "one of args for show_vedo is neither `dict` nor",
                    "`list`. Putting it naively into a list."
            )
            showlist = [arg]

        # quickcheck if the list is gustaf or non-gustaf
        # if gustaf, make it vedo-showable.
        # if there's spline, we need to pop the element and
        # extend showables to the list.
        # A showlist is a list to be plotted into a single subframe of the
        # plot
        list_of_showables = []
        for sl in showlist:
            if not isinstance(sl, list):
                sl = [sl]
            for k, item in enumerate(sl):
                if isinstance(item, GustafBase):
                    tmp_showable = item.showable(backend="vedo", **kwargs)
                    # splines return dict
                    # - maybe it is time to do some typing..
                    if isinstance(tmp_showable, dict):
                        # add to extend later
                        list_of_showables.extend(list(tmp_showable.values()))

                    else:
                        # replace gustafobj with vedo_obj.
                        list_of_showables.append(tmp_showable)
                else:
                    list_of_showables.extend(sl)

        # set interactive to true at last element
        if int(i + 1) == len(args):
            plt.show(
                    list_of_showables,
                    at=i,
                    interactive=interac,
                    camera=cam_tuple_to_list(cam),
                    # offscreen=offs,
            )

        else:
            plt.show(
                    list_of_showables,
                    at=i,
                    interactive=False,
                    camera=cam_tuple_to_list(cam),
                    # offscreen=offs,
            )

    if interac and not offs:
        # only way to ensure memory is released
        clear_vedoplotter(plt, np.prod(plt.shape))

        if close or close is None:  # explicitly given or None.
            # It seems to leak some memory, but here it goes.
            plt.close()  # if i close it, this cannot be reused...
            plt = None

    if return_show_list:
        return (plt, list_of_showables)
    else:
        return plt


def _vedo_showable(obj, **kwargs):
    """Generates a vedo obj based on `kind` attribute from given obj.

    Parameters
    -----------
    obj: gustaf obj

    Returns
    --------
    vedo_obj: vedo obj
    """
    # parse from vis_dict
    # NOTE: maybe we can make a helper class to organize this nicely
    basic_options = dict(
            c=obj.vis_dict.get("c", None),
            r=obj.vis_dict.get("r", None),
            lw=obj.vis_dict.get("lw", None),
            alpha=obj.vis_dict.get("alpha", None),
            cmap=obj.vis_dict.get("cmap", None),
            # > followings are cmap options
            vmin=obj.vis_dict.get("vmin", None),
            vmax=obj.vis_dict.get("vmax", None),
            cmapalpha=obj.vis_dict.get("cmapalpha", 1),
            # > takes scalarbar options as dict.
            scalarbar=obj.vis_dict.get("scalarbar", None),
            dataname=obj.vis_dict.get("dataname", None),
            # <
            arrows=obj.vis_dict.get("arrows", None),  # only for edges
            # >only for edges internally treated same as `lw`,
            # but higher priority
            thickness=obj.vis_dict.get("thickness", None),
            title=obj.vis_dict.get("title", "gustaf")
    )
    # loop once more to extract basics from kwargs
    # done after vis_dict, so that this overpowers
    keys = list(kwargs.keys())  # to pop dict during loop
    for key in keys:
        if key in basic_options.keys():
            basic_options[key] = kwargs[key]
            kwargs.pop(key)

    utils.log.debug("making vedo-showable obj")
    local_options = dict()

    if obj.kind == "vertex":
        for key in ["c", "r", "alpha"]:
            value = basic_options[key]
            if value is not None:
                local_options.update({key: value})

        vobj = vedo.Points(
                obj.vertices,
                **local_options,
        )

    elif obj.kind == "edge":
        for key in ["c", "lw", "alpha"]:
            value = basic_options[key]
            if value is not None:
                local_options.update({key: value})

        # edges can be arrows if vis_dict["arrows"] is set True
        if not basic_options["arrows"]:
            vobj = vedo.Lines(
                    obj.vertices[obj.edges],
                    **local_options,
            )

        else:
            if basic_options.get("thickness", False):
                local_options.update({"thickness": basic_options["thickness"]})

            # turn lw into thickness if there's no thickness
            elif local_options.get("lw", False):
                thickness = local_options.pop("lw")
                local_options.update({"thickness": thickness})

            # `s` is another param for arrows
            local_options.update({"s": obj.vis_dict.get("s", None)})

            vobj = vedo.Arrows(
                    obj.vertices[obj.edges],
                    **local_options,
            )

    elif obj.kind == "face":
        for key in ["c", "alpha"]:
            value = basic_options[key]
            if value is not None:
                local_options.update({key: value})

        vobj = vedo.Mesh(
                [obj.vertices, obj.faces],
                **local_options,
        )

    elif obj.kind == "volume":
        from vtk import VTK_TETRA as frau_tetra
        from vtk import VTK_HEXAHEDRON as herr_hexa

        whatami = obj.whatami
        if whatami.startswith("tet"):
            grid_type = frau_tetra
        elif whatami.startswith("hexa"):
            grid_type = herr_hexa
        else:
            return None  # get_whatami should've rasied error..

        if basic_options["dataname"]:
            from gustaf.faces import Faces

            # UGrid would be politically correct,
            # but currently it can't show field
            # so, extract only surface mesh
            surf_ids = obj.single_faces()  # gets faces too
            sfaces = Faces(obj.vertices, obj.faces()[surf_ids])
            sfaces.remove_unreferenced_vertices()

            vobj = sfaces.showable(backend="vedo")  # recursive alert

        else:
            vobj = vedo.UGrid(
                    [
                            obj.vertices,
                            obj.volumes,
                            np.repeat([grid_type], len(obj.volumes)),
                    ]
            )

            if basic_options["c"] is None:
                basic_options["c"] = "hotpink"

            vobj.color(basic_options["c"])
            vobj.alpha(basic_options["alpha"])

    # this sets vedo v2021.0.6+ requirement
    dname = basic_options["dataname"]
    if dname is not None:
        # transfer data
        vobj.pointdata[dname] = obj.vertexdata[dname]

        # default cmap is jet.
        if basic_options["cmap"] is None:
            basic_options["cmap"] = "jet"

        # register cmap and data
        vobj.cmap(
                basic_options["cmap"],
                input_array=dname,
                on="points",  # hardcoded since yet, we don't have cell field
                vmin=basic_options["vmin"],
                vmax=basic_options["vmax"],
                alpha=basic_options["cmapalpha"],
        )

        # scalarbar?
        scalarbar_dict = basic_options["scalarbar"]
        if scalarbar_dict is not None:
            # if horizontal==True, size doesnt really matter
            vobj.addScalarBar(**scalarbar_dict)

    return vobj


def _trimesh_showable(obj):
    """"""
    pass


def _matplotlib_showable(obj):
    """"""
    pass


def make_showable(obj, backend=settings.VISUALIZATION_BACKEND, **kwargs):
    """Since gustaf does not natively support visualization, one of the
    following library is used to visualize gustaf (visualizable) objects: (1)
    vedo -> Fast, offers a lot of features (2) trimesh -> Fast, compatible with
    old opengl (3) matplotlib -> Slow, offers vector graphics.

    This determines showing types using `whatami`.

    Parameters
    -----------
    obj: gustaf-objects
    backend: str
      (Optional) Default is `gustaf.settings.VISUALIZATION_BACKEND`.
      Options are: "vedo" | "trimesh" | "matplotlib"

    Returns
    --------
    showalbe_objs: list
      List of showable objects.
    """
    if backend.startswith("vedo"):
        return _vedo_showable(obj, **kwargs)
    elif backend.startswith("trimesh"):
        return _trimesh_showable(obj, **kwargs)
    elif backend.startswith("matplotlib"):
        return _matplotlib_showable(obj, **kwargs)
    else:
        raise NotImplementedError


# possibly relocate
def interpolate_vedo_dictcam(cameras, resolutions, spline_degree=1):
    """Interpolate between vedo dict cameras.

    Parameters
    ------------
    cameras: list or tuple
    resolutions: int
    spline_degree: int
      if > 1 and splinepy is available and there are more than two cameras,
      we interpolate all the entries using spline.

    Returns
    --------
    interpolated_cams: list
    """
    try:
        import splinepy
        spp = True

    except ImportError:
        spp = False

    # quick type check loop
    camkeys = ["pos", "focalPoint", "viewup", "distance", "clippingRange"]
    for cam in cameras:
        if not isinstance(cam, dict):
            raise TypeError("Only `dict` description of vedo cam is allowed.")
        else:
            for key in camkeys:
                if cam[key] is None:
                    raise ValueError(
                            f"One of the camera does not contain `{key}` info"
                    )

    interpolated_cams = []
    total_cams = int(resolutions) * (len(cameras) - 1)

    if spp and spline_degree > 1 and len(cameras) > 2:
        if spline_degree > len(cameras):
            raise ValueError(
                    "Not enough camera to interpolate with "
                    f"spline degree {spline_degree}"
            )

        ps = []
        fs = []
        vs = []
        ds = []
        cs = []
        for cam in cameras:
            ps.append(list(cam[camkeys[0]]))
            fs.append(list(cam[camkeys[1]]))
            vs.append(list(cam[camkeys[2]]))
            ds.append([float(cam[camkeys[3]])])
            cs.append(list(cam[camkeys[4]]))

        interpolated = dict()
        for i, prop in enumerate([ps, fs, vs, ds, cs]):
            ispline = splinepy.BSpline()
            ispline.interpolate_curve(
                    query_points=prop,
                    degree=spline_degree,
                    save_query=False,
            )
            interpolated[camkeys[i]] = ispline.sample([total_cams])

        for i in range(total_cams):
            interpolated_cams.append(
                    {
                            camkeys[0]: interpolated[camkeys[0]][i].tolist(),
                            camkeys[1]: interpolated[camkeys[1]][i].tolist(),
                            camkeys[2]: interpolated[camkeys[2]][i].tolist(),
                            camkeys[3]:
                            interpolated[camkeys[3]][i][0],  # float?
                            camkeys[4]: interpolated[camkeys[4]][i].tolist(),
                    }
            )

    else:
        i = 0
        for startcam, endcam in zip(cameras[:-1], cameras[1:]):
            if i == 0:
                interpolated = [
                        np.linspace(
                                startcam[ckeys],
                                endcam[ckeys],
                                resolutions,
                        ).tolist() for ckeys in camkeys
                ]

            else:
                interpolated = [
                        np.linspace(
                                startcam[ckeys],
                                endcam[ckeys],
                                int(resolutions + 1),
                        )[1:].tolist() for ckeys in camkeys
                ]

            i += 1

            for j in range(resolutions):
                interpolated_cams.append(
                        {
                                camkeys[0]: interpolated[0][j],
                                camkeys[1]: interpolated[1][j],
                                camkeys[2]: interpolated[2][j],
                                camkeys[3]: interpolated[3][j],  # float?
                                camkeys[4]: interpolated[4][j],
                        }
                )

    return interpolated_cams



"""gustaf/gustaf/_base.py.

Useful base class for gustaf. Plus some useful decorators.
"""
from gustaf import utils


class GustafBase:
    """Base class for gustaf, where logging is nicely wrapped, and some useful
    methods are defined as classmethods..

    Since attributes are predefined with __slots__, we can pre define
    all the properties that could have been saved.
    Adding `get_` in front of such properties are names for functions that
    freshly compute and save the properties and their byproducts.

    Other more complex operations will be a separate function.

    TODO: maybe add explicit `use_saved` switch to avoid recomputing
    """

    __slots__ = ()

    def _logd(self, *log):
        """Debug logger wrapper for Mesh.

        Parameters
        -----------
        *log: *str

        Returns
        --------
        None
        """
        utils.log.debug(type(self).__qualname__, "-", *log)

    def _logi(self, *log):
        """Info logger wrapper for Mesh.

        Parameters
        -----------
        *log: *str

        Returns
        --------
        None
        """
        utils.log.info(type(self).__qualname__, "-", *log)

    def _logw(self, *log):
        """Warning logger wrapper for Mesh.

        Parameters
        -----------
        *log: *str

        Returns
        --------
        None
        """
        utils.log.warning(type(self).__qualname__, "-", *log)



"""gustaf/spline/_utils.py.

Utils for spline. Internal use only.
"""


def to_res_list(res, length):
    """Given int or list, returns a list of resolutions of correct length. If
    list is given, checks if it is a correct length.

    Parameters
    -----------
    res: int or list
    length: int
      length of desired res_list. Usually corresponds to spline.para_dim

    Returns
    --------
    res_list: list
    """
    if isinstance(res, list):
        if len(res) != length:
            raise ValueError(
                    "Invalid resolution length! " + "It should match length."
            )

        return res

    elif isinstance(res, (int, float)):
        res_list = [res for _ in range(length)]
        return res_list

    elif isinstance(res, tuple):
        return to_res_list(list(res), length)

    else:
        raise TypeError(
                "Invalid resolutions input. It should be int, tuple, or list."
        )



"""gustaf/spline/proximity.py.

Closest what?
"""


def closest_control_points(
        spline,
        query_points,
        return_distances=False,
):
    """Returns indices of closest control points.

    Parameters
    -----------
    spline: BSpline or NURBS
    query_points: (n, spline.dim) np.ndarray
      float
    return_distances: bool

    tolerance: float
      Default is settings.Tolerance. Only relevant iff nearest_only==False

    Returns
    --------
    indices: (n,) np.ndarray
    distances: (n, spline.dim) np.ndarray
      iff return_distances==True
    """
    from scipy.spatial import cKDTree as KDTree

    kdt = KDTree(spline.control_points)

    dist, ids = kdt.query(query_points)

    if return_distances:
        return ids, dist

    else:
        return ids


def closest_parametric_coordinate(
        spline,
        query_points,
):
    """Finds closest points using."""
    pass


class Proximity:
    """Helper class to allow direct proximity queries for spline obj (BSpline
    or NURBS). Internal use only.

    Examples
    ---------
    >>> myspline = <your-spline>
    >>> closest_cp_ids = myspline.proximity.closest_control_points(queries)
    """

    def __init__(self, spl):
        self.spline = spl

    def closest_control_points(self, *args, **kwargs):
        return closest_control_points(self.spline, *args, **kwargs)



import numpy as np

from gustaf.spline import base


class CrossTile3D(base.GustafBase):

    def __init__(self):
        """Simple crosstile with linear-quadratic branches and a trilinear
        center spline."""
        self._dim = 3
        self._evaluation_points = np.array(
                [
                        [0., .5, .5],
                        [1., .5, .5],
                        [.5, 0., .5],
                        [.5, 1., .5],
                        [.5, .5, 0.],
                        [.5, .5, 1.],
                ]
        )
        self._parameter_space_dimension = 1

    @property
    def parameter_space_dimension(self):
        """Number of parameters per evaluation point."""
        return self._parameter_space_dimension

    @property
    def evaluation_points(self):
        """Positions in the parametrization function to be evaluated when tile
        " "is constructed prior to composition.

        Parameters
        ----------
        None

        Returns
        -------
        evaluation_points : np.ndarray(6,3)
        """
        return self._evaluation_points

    @property
    def dim(self):
        """Returns dimensionality in physical space of the Microtile.

        Parameters
        ----------
        None

        Returns
        -------
        dim : int
        """
        return self._dim

    def closing_tile(
            self,
            parameters=None,
            closure=None,
            boundary_width=0.1,
            filling_height=0.5,
            **kwargs
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters : tuple(np.ndarray)
          radii of fitting cylinder at evaluation points
        closure : str
          parametric dimension that needs to be closed.
          Must be {"z_min", "z_max"}
        boundary_width : float
          with of the boundary surronding branch
        filling_height : float
          portion of the height that is filled in parametric domain

        Results
        -------
        list_of_splines : list
        """
        # Check parameters
        if closure is None:
            raise ValueError("No closing direction given")

        if parameters is None:
            self._logd("Tile request is not parametrized, setting default 0.2")
            parameters = tuple([np.ones(6) * 0.2])
        parameters = parameters[0]
        if not (np.all(parameters > 0) and np.all(parameters < .5)):
            raise ValueError("Thickness out of range (0, .5)")

        if not (0. < float(boundary_width) < .5):
            raise ValueError("Boundary Width is out of range")

        if not (0. < float(filling_height) < 1.):
            raise ValueError("Filling must  be in (0,1)")

        inv_boundary_width = 1. - boundary_width
        inv_filling_height = 1. - filling_height
        center_width = 1. - 2 * boundary_width
        ctps_mid_height_top = (1 + filling_height) * .5
        ctps_mid_height_bottom = 1. - ctps_mid_height_top
        r_center = center_width * .5

        spline_list = []
        if closure == "z_min":
            # The branch is located at zmin of current tile
            branch_thickness = parameters[5]
            ctps_corner = np.array(
                    [
                            [0., 0., 0.], [boundary_width, 0., 0.],
                            [0., boundary_width, 0.],
                            [boundary_width, boundary_width, 0.],
                            [0., 0., filling_height],
                            [boundary_width, 0., filling_height],
                            [0., boundary_width, filling_height],
                            [boundary_width, boundary_width, filling_height]
                    ]
            )

            spline_list.append(
                    base.Bezier(degrees=[1, 1, 1], control_points=ctps_corner)
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=(
                                    ctps_corner
                                    + np.array([0., inv_boundary_width, 0.])
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=(
                                    ctps_corner
                                    + np.array([inv_boundary_width, 0., 0.])
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=(
                                    ctps_corner + np.array(
                                            [
                                                    inv_boundary_width,
                                                    inv_boundary_width, 0.
                                            ]
                                    )
                            )
                    )
            )

            center_ctps = np.array(
                    [
                            [boundary_width, boundary_width, 0.],
                            [inv_boundary_width, boundary_width, 0.],
                            [boundary_width, inv_boundary_width, 0.],
                            [inv_boundary_width, inv_boundary_width, 0.],
                            [boundary_width, boundary_width, filling_height],
                            [
                                    inv_boundary_width, boundary_width,
                                    filling_height
                            ],
                            [
                                    boundary_width, inv_boundary_width,
                                    filling_height
                            ],
                            [
                                    inv_boundary_width, inv_boundary_width,
                                    filling_height
                            ]
                    ]
            )

            spline_list.append(
                    base.Bezier(degrees=[1, 1, 1], control_points=center_ctps)
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=np.maximum(
                                    center_ctps
                                    - np.array([center_width, 0, 0]), 0
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=np.maximum(
                                    center_ctps
                                    - np.array([0, center_width, 0]), 0
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=np.minimum(
                                    center_ctps
                                    + np.array([center_width, 0, 0]), 1.
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=np.minimum(
                                    center_ctps
                                    + np.array([0, center_width, 0]), 1.
                            )
                    )
            )
            branch_ctps = np.array(
                    [
                            [-r_center, -r_center, filling_height],
                            [r_center, -r_center, filling_height],
                            [-r_center, r_center, filling_height],
                            [r_center, r_center, filling_height],
                            [
                                    -branch_thickness, -branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    branch_thickness, -branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    -branch_thickness, branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    branch_thickness, branch_thickness,
                                    ctps_mid_height_top
                            ], [-branch_thickness, -branch_thickness, 1.],
                            [branch_thickness, -branch_thickness, 1.],
                            [-branch_thickness, branch_thickness, 1.],
                            [branch_thickness, branch_thickness, 1.]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(degrees=[1, 1, 2], control_points=branch_ctps)
            )

            return spline_list
        elif closure == "z_max":
            # The branch is located at zmax of current tile
            branch_thickness = parameters[4]
            ctps_corner = np.array(
                    [
                            [0., 0., inv_filling_height],
                            [boundary_width, 0., inv_filling_height],
                            [0., boundary_width, inv_filling_height],
                            [
                                    boundary_width, boundary_width,
                                    inv_filling_height
                            ], [0., 0., 1.], [boundary_width, 0., 1.],
                            [0., boundary_width, 1.],
                            [boundary_width, boundary_width, 1.]
                    ]
            )

            spline_list.append(
                    base.Bezier(degrees=[1, 1, 1], control_points=ctps_corner)
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=(
                                    ctps_corner
                                    + np.array([0., inv_boundary_width, 0.])
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=(
                                    ctps_corner
                                    + np.array([inv_boundary_width, 0., 0.])
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=(
                                    ctps_corner + np.array(
                                            [
                                                    inv_boundary_width,
                                                    inv_boundary_width, 0.
                                            ]
                                    )
                            )
                    )
            )

            center_ctps = np.array(
                    [
                            [
                                    boundary_width, boundary_width,
                                    inv_filling_height
                            ],
                            [
                                    inv_boundary_width, boundary_width,
                                    inv_filling_height
                            ],
                            [
                                    boundary_width, inv_boundary_width,
                                    inv_filling_height
                            ],
                            [
                                    inv_boundary_width, inv_boundary_width,
                                    inv_filling_height
                            ], [boundary_width, boundary_width, 1.],
                            [inv_boundary_width, boundary_width, 1.],
                            [boundary_width, inv_boundary_width, 1.],
                            [inv_boundary_width, inv_boundary_width, 1.]
                    ]
            )

            spline_list.append(
                    base.Bezier(degrees=[1, 1, 1], control_points=center_ctps)
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=np.maximum(
                                    center_ctps
                                    - np.array([center_width, 0, 0]), 0
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=np.maximum(
                                    center_ctps
                                    - np.array([0, center_width, 0]), 0
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=np.minimum(
                                    center_ctps
                                    + np.array([center_width, 0, 0]), 1.
                            )
                    )
            )

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 1, 1],
                            control_points=np.minimum(
                                    center_ctps
                                    + np.array([0, center_width, 0]), 1.
                            )
                    )
            )

            branch_ctps = np.array(
                    [
                            [-branch_thickness, -branch_thickness, 0.],
                            [branch_thickness, -branch_thickness, 0.],
                            [-branch_thickness, branch_thickness, 0.],
                            [branch_thickness, branch_thickness, 0.],
                            [
                                    -branch_thickness, -branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    branch_thickness, -branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    -branch_thickness, branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    branch_thickness, branch_thickness,
                                    ctps_mid_height_bottom
                            ], [-r_center, -r_center, inv_filling_height],
                            [r_center, -r_center, inv_filling_height],
                            [-r_center, r_center, inv_filling_height],
                            [r_center, r_center, inv_filling_height]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(degrees=[1, 1, 2], control_points=branch_ctps)
            )

            return spline_list
        else:
            raise NotImplementedError(
                    "Requested closing dimension is not supported"
            )

    def create_tile(self, parameters=None, center_expansion=1., **kwargs):
        """Create a microtile based on the parameters that describe the branch
        thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : tuple(np.array)
          only first entry is used, defines the internal radii of the
          branches
        center_expansion : float
          thickness of center is expanded by a factor

        Returns
        -------
        microtile_list : list(splines)
        """

        if not isinstance(center_expansion, float):
            raise ValueError("Invalid Type")
        if not ((center_expansion > .5) and (center_expansion < 1.5)):
            raise ValueError("Center Expansion must be in (.5,1.5)")
        max_radius = min(.5, (.5 / center_expansion))
        # set to default if nothing is given
        if parameters is None:
            self._logd("Setting branch thickness to default 0.2")
            parameters = tuple([np.ones(6) * 0.2])
        [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r,
         z_max_r] = parameters[0].tolist()
        for radius in [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r, z_max_r]:
            if not isinstance(radius, float):
                raise ValueError("Invalid type")
            if not (radius > 0 and radius < max_radius):
                raise ValueError(
                        f"Radii must be in (0,{max_radius}) for "
                        f"center_expansion {center_expansion}"
                )

        # center radius
        center_r = (
                x_min_r + x_max_r + y_min_r + y_max_r + z_min_r + z_max_r
        ) / 6. * center_expansion
        hd_center = 0.5 * (0.5 + center_r)

        # Create the center-tile
        center_points = np.array(
                [
                        [-center_r, -center_r, -center_r],
                        [center_r, -center_r, -center_r],
                        [-center_r, center_r, -center_r],
                        [center_r, center_r, -center_r],
                        [-center_r, -center_r, center_r],
                        [center_r, -center_r, center_r],
                        [-center_r, center_r, center_r],
                        [center_r, center_r, center_r]
                ]
        )

        center_spline = base.Bezier(
                degrees=[1, 1, 1], control_points=center_points + [.5, .5, .5]
        )

        # X-Axis branches
        # X-Min-Branch
        aux_x_min = min(x_min_r, center_r)
        x_min_ctps = np.array(
                [
                        [-.5, -x_min_r, -x_min_r],
                        [-hd_center, -aux_x_min,
                         -aux_x_min], center_points[0, :],
                        [-.5, x_min_r, -x_min_r],
                        [-hd_center, aux_x_min,
                         -aux_x_min], center_points[2, :],
                        [-.5, -x_min_r, x_min_r],
                        [-hd_center, -aux_x_min, aux_x_min],
                        center_points[4, :], [-.5, x_min_r, x_min_r],
                        [-hd_center, aux_x_min, aux_x_min], center_points[6, :]
                ]
        )
        x_min_spline = base.Bezier(
                degrees=[2, 1, 1], control_points=x_min_ctps + [.5, .5, .5]
        )
        # X-Min-Branch
        aux_x_max = min(x_max_r, center_r)
        x_max_ctps = np.array(
                [
                        center_points[1, :],
                        [hd_center, -aux_x_max, -aux_x_max],
                        [.5, -x_max_r, -x_max_r], center_points[3, :],
                        [hd_center, aux_x_max, -aux_x_max],
                        [.5, x_max_r, -x_max_r], center_points[5, :],
                        [hd_center, -aux_x_max, aux_x_max],
                        [.5, -x_max_r, x_max_r], center_points[7, :],
                        [hd_center, aux_x_max, aux_x_max],
                        [.5, x_max_r, x_max_r]
                ]
        )
        x_max_spline = base.Bezier(
                degrees=[2, 1, 1], control_points=x_max_ctps + [.5, .5, .5]
        )

        # Y-Axis branches
        # Y-Min-Branch
        aux_y_min = min(y_min_r, center_r)
        y_min_ctps = np.array(
                [
                        [-y_min_r, -.5, -y_min_r], [y_min_r, -.5, -y_min_r],
                        [-aux_y_min, -hd_center, -aux_y_min],
                        [aux_y_min, -hd_center,
                         -aux_y_min], center_points[0, :], center_points[1, :],
                        [-y_min_r, -.5, y_min_r], [y_min_r, -.5, y_min_r],
                        [-aux_y_min, -hd_center, aux_y_min],
                        [aux_y_min, -hd_center, aux_y_min],
                        center_points[4, :], center_points[5, :]
                ]
        )
        y_min_spline = base.Bezier(
                degrees=[1, 2, 1], control_points=y_min_ctps + [.5, .5, .5]
        )
        # Y-Min-Branch
        aux_y_max = min(y_max_r, center_r)
        y_max_ctps = np.array(
                [
                        center_points[2, :], center_points[3, :],
                        [-aux_y_max, hd_center, -aux_y_max],
                        [aux_y_max, hd_center, -aux_y_max],
                        [-y_max_r, .5, -y_max_r], [y_max_r, .5, -y_max_r],
                        center_points[6, :], center_points[7, :],
                        [-aux_y_max, hd_center, aux_y_max],
                        [aux_y_max, hd_center, aux_y_max],
                        [-y_max_r, .5, y_max_r], [y_max_r, .5, y_max_r]
                ]
        )
        y_max_spline = base.Bezier(
                degrees=[1, 2, 1], control_points=y_max_ctps + [.5, .5, .5]
        )

        # Y-Axis branches
        # Y-Min-Branch
        aux_z_min = min(z_min_r, center_r)
        z_min_ctps = np.array(
                [
                        [-z_min_r, -z_min_r, -.5], [z_min_r, -z_min_r, -.5],
                        [-z_min_r, z_min_r, -.5], [z_min_r, z_min_r, -.5],
                        [-aux_z_min, -aux_z_min, -hd_center],
                        [aux_z_min, -aux_z_min, -hd_center],
                        [-aux_z_min, aux_z_min, -hd_center],
                        [aux_z_min, aux_z_min,
                         -hd_center], center_points[0, :], center_points[1, :],
                        center_points[2, :], center_points[3, :]
                ]
        )
        z_min_spline = base.Bezier(
                degrees=[1, 1, 2], control_points=z_min_ctps + [.5, .5, .5]
        )
        # Y-Min-Branch
        aux_z_max = min(z_max_r, center_r)
        z_max_ctps = np.array(
                [
                        center_points[4, :], center_points[5, :],
                        center_points[6, :], center_points[7, :],
                        [-aux_z_max, -aux_z_max, hd_center],
                        [aux_z_max, -aux_z_max, hd_center],
                        [-aux_z_max, aux_z_max, hd_center],
                        [aux_z_max, aux_z_max, hd_center],
                        [-z_max_r, -z_max_r, .5], [z_max_r, -z_max_r, .5],
                        [-z_max_r, z_max_r, .5], [z_max_r, z_max_r, .5]
                ]
        )
        z_max_spline = base.Bezier(
                degrees=[1, 1, 2], control_points=z_max_ctps + [.5, .5, .5]
        )

        return [
                center_spline, x_min_spline, x_max_spline, y_min_spline,
                y_max_spline, z_min_spline, z_max_spline
        ]



from gustaf import _version
from gustaf import settings
from gustaf import vertices
from gustaf import edges
from gustaf import faces
from gustaf import volumes
from gustaf import show
from gustaf import utils
from gustaf import create
from gustaf import io
from gustaf import helpers
from gustaf.vertices import Vertices
from gustaf.edges import Edges
from gustaf.faces import Faces
from gustaf.volumes import Volumes

has_spline = False
try:
    from gustaf import spline
    from gustaf.spline.base import BSpline, NURBS, Bezier, RationalBezier
    from gustaf.spline.ffd import FFD
    has_spline = True
except ImportError as err:
    # overwrites the all modules which depend on the `splinepy` library
    # with an object which will throw an error
    # as soon as it is used the first time. This means that any non spline
    # functionality works as before, but as soon as these are used a
    # comprehensive exception will be raised which is understandable in
    # contrast to the possible multitude of errors previously possible
    from gustaf.helpers.raise_if import ModuleImportRaiser
    spline = ModuleImportRaiser("splinepy", err)
    BSpline = spline
    NURBS = spline
    Bezier = spline
    RationalBezier = spline
    FFD = spline

# import try/catch for triangle and gustaf-tetgen

__version__ = _version.version

__all__ = [
        "__version__",
        "settings",
        "vertices",
        "edges",
        "faces",
        "volumes",
        "show",
        "utils",
        "create",
        "io",
        "helpers",
        "Vertices",
        "Edges",
        "Faces",
        "Volumes",
        "spline",
        "has_spline",
        "BSpline",
        "NURBS",
        "Bezier",
        "RationalBezier",
        "FFD",
]



"""gustaf/gustaf/helpers/data.py.

Helps helpee to manage data. Some useful data structures.
"""

import abc
from functools import wraps
from collections import namedtuple

import numpy as np


class TrackedArray(np.ndarray):
    """Taken from nice implementations of `trimesh` (see LICENSE.txt).
    `https://github.com/mikedh/trimesh/blob/main/trimesh/caching.py`. Minor
    adaption, since we don't have hashing functionalities.

    All the inplace functions will set modified flag and if some operations
    has potential to cause un-trackable behavior, writeable flags will be set
    to False.

    Note, if you really really want, it is possible to change the tracked
    array without setting modified flag.
    """

    __slots__ = ("_modified", "_source")

    def __array_finalize__(self, obj):
        """Sets default flags for any arrays that maybe generated based on
        tracked array."""
        self._modified = True
        self._source = int(0)

        if isinstance(obj, type(self)):
            if isinstance(obj._source, int):
                self._source = obj
            else:
                self._source = obj._source

    @property
    def mutable(self):
        return self.flags['WRITEABLE']

    @mutable.setter
    def mutable(self, value):
        self.flags.writeable = value

    def _set_modified(self):
        """set modified flags to itself and to the source."""
        self._modified = True
        if isinstance(self._source, type(self)):
            self._source._modified = True

    def copy(self, *args, **kwargs):
        """copy gives np.ndarray.

        no more tracking.
        """
        return np.array(self, copy=True)

    def view(self, *args, **kwargs):
        """Set writeable flags to False for the view."""
        v = super(self.__class__, self).view(*args, **kwargs)
        v.flags.writeable = False
        return v

    def __iadd__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__iadd__(*args, **kwargs)

    def __isub__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__isub__(*args, **kwargs)

    def __imul__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__imul__(*args, **kwargs)

    def __idiv__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__idiv__(*args, **kwargs)

    def __itruediv__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__itruediv__(*args, **kwargs)

    def __imatmul__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__imatmul__(*args, **kwargs)

    def __ipow__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ipow__(*args, **kwargs)

    def __imod__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__imod__(*args, **kwargs)

    def __ifloordiv__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ifloordiv__(*args, **kwargs)

    def __ilshift__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ilshift__(*args, **kwargs)

    def __irshift__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__irshift__(*args, **kwargs)

    def __iand__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__iand__(*args, **kwargs)

    def __ixor__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ixor__(*args, **kwargs)

    def __ior__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ior__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        self._set_modified()
        super(self.__class__, self).__setitem__(*args, **kwargs)

    def __setslice__(self, *args, **kwargs):
        self._set_modified()
        super(self.__class__, self).__setslice__(*args, **kwargs)

    def __getslice__(self, *args, **kwargs):
        self._set_modified()
        """
        return slices I am pretty sure np.ndarray does not have __*slice__
        """
        slices = super(self.__class__, self).__getitem__(*args, **kwargs)
        if isinstance(slices, np.ndarray):
            slices.flags.writeable = False
        return slices


def make_tracked_array(array, dtype=None, copy=True):
    """Taken from nice implementations of `trimesh` (see LICENSE.txt).
    `https://github.com/mikedh/trimesh/blob/main/trimesh/caching.py`.

    ``Properly subclass a numpy ndarray to track changes.
    Avoids some pitfalls of subclassing by forcing contiguous
    arrays and does a view into a TrackedArray.``

    Factory-like wrapper function for TrackedArray.

    Parameters
    ------------
    array: array- like object
      To be turned into a TrackedArray
    dtype: np.dtype
      Which dtype to use for the array
    copy: bool
      Default is True. copy if True.

    Returns
    ------------
    tracked : TrackedArray
      Contains input array data
    """
    # if someone passed us None, just create an empty array
    if array is None:
        array = []
    # make sure it is contiguous then view it as our subclass
    tracked = np.ascontiguousarray(array, dtype=dtype)
    if copy:
        tracked = tracked.copy().view(TrackedArray)
    else:
        tracked = tracked.view(TrackedArray)

    # should always be contiguous here
    assert tracked.flags['C_CONTIGUOUS']

    return tracked


class DataHolder(abc.ABC):
    __slots__ = [
            "_helpee",
            "_saved",
    ]

    def __init__(self, helpee):
        """Base class for any data holder. Behaves similar to dict.

        Attributes
        -----------
        None

        Parameters
        -----------
        helpee: object
          GustafBase objects would probably make the most sense here.
        """
        self._helpee = helpee
        self._saved = dict()

    def __setitem__(self, key, value):
        """Raise Error to disable direct value setting.

        Parameters
        -----------
        key: str
        value: object

        Returns
        --------
        None
        """
        raise NotImplementedError(
                "Sorry, you can't set items directly for "
                f"{type(self).__qualname__}"
        )

    def __getitem__(self, key):
        """Returns stored item if the key exists.

        Parameters
        -----------
        key: str

        Returns
        --------
        value: object
        """
        if key in self._saved.keys():
            return self._saved[key]

        else:
            raise KeyError(f"`{key}` is not stored for {type(self._helpee)}")

    def get(self, key, default_values=None):
        """Returns stored item if the key exists. Else, given default value. If
        the key exist, default value always exists, since it is initialized
        that way.

        Parameters
        -----------
        key: str
        default_values: object

        Returns
        --------
        value: np.ndarray
        """
        if key in self._saved.keys():
            return self._saved[key]
        else:
            return default_values

    def keys(self):
        """Returns keys of data holding dict.

        Parameters
        -----------
        None

        Returns
        --------
        keys: dict_keys
        """
        return self._saved.keys()

    def values(self):
        """Returns values of data holding dict.

        Parameters
        -----------
        None

        Returns
        --------
        values: dict_values
        """
        return self._saved.values()

    def items(self):
        """Returns items of data holding dict.

        Parameters
        -----------
        None

        Returns
        --------
        values: dict_values
        """
        return self._saved.items()


class ComputedData(DataHolder):
    _depends = None
    _inv_depends = None

    ___slots___ = []

    def __init__(self, helpee, **kwargs):
        """Stores last computed values. Keys are expected to be the same as
        helpee's function that computes the value.

        Attributes
        -----------
        None

        Parameters
        -----------
        helpee: GustafBase
        """
        super().__init__(helpee)

    @classmethod
    def depends_on(cls, var_names, make_property=False):
        """Decorator as classmethod.

        checks if the key should be computed. Two cases, where the answer is
        yes:

        1. there's modification on arrays that the key depend on.
            ->erases all other
        2. is corresponding value None?

        Supports multi-dependency

        Parameters
        -----------
        var_name: list
        make_property:
        """

        def inner(func):
            # followings are done once while modules are loaded
            # just subclass this class to make a special helper
            # for each helpee class.
            assert isinstance(var_names, list), "var_names should be a list"
            # initialize property
            # _depends is dict(str: list)
            if cls._depends is None:
                cls._depends = dict()
            if cls._depends.get(func.__name__, None) is None:
                cls._depends[func.__name__] = list()
            # add dependency info
            cls._depends[func.__name__].extend(var_names)

            # _inv_depends is dict(str: list)
            if cls._inv_depends is None:
                cls._inv_depends = dict()
            # add inverse dependency
            for vn in var_names:
                if cls._inv_depends.get(vn, None) is None:
                    cls._inv_depends[vn] = list()

                cls._inv_depends[vn].append(func.__name__)

            @wraps(func)
            def compute_or_return_saved(*args, **kwargs):
                """Check if the key should be computed,"""
                # extract some related info
                self = args[0]  # the helpee itself
                recompute = kwargs.get("recompute", False)
                return_saved = kwargs.get("return_saved", False)

                # if return_saved, try to escape as soon as possible
                if return_saved:
                    saved = self._computed._saved.get(func.__name__, None)
                    if saved is not None and not recompute:
                        return saved

                # computed arrays are called _computed.
                # loop over dependees and check if they are modified
                for dependee_str in cls._depends[func.__name__]:
                    dependee = getattr(self, dependee_str)
                    # is modified?
                    if dependee._modified:
                        for inv in cls._inv_depends[dependee_str]:
                            self._computed._saved[inv] = None

                # is saved / want to recompute?
                # recompute is added for computed values that accepts params.
                saved = self._computed._saved.get(func.__name__, None)
                if saved is not None and not recompute:
                    return saved

                # we've reached this point because we have to compute this
                computed = func(*args, **kwargs)
                if isinstance(computed, np.ndarray):
                    computed.flags.writeable = False  # configurable?
                self._computed._saved[func.__name__] = computed

                # so, all fresh. we can press NOT-modified  button
                for dependee_str in cls._depends[func.__name__]:
                    dependee = getattr(self, dependee_str)
                    dependee._modified = False

                return computed

            if make_property:
                return property(compute_or_return_saved)
            else:
                return compute_or_return_saved

        return inner


Unique2DFloats = namedtuple(
        "Unique2DFloats", ["values", "ids", "inverse", "intersection"]
)
"""
namedtuple to hold unique information of float type arrays.
Note that for float types, "close enough" might be a better name than unique.
This way, all tracked arrays, as long as they are 2D, have a dot separated
syntax to access unique info. For example, `mesh.unique_vertices.ids`.
"""
Unique2DFloats.values.__doc__ = """`(n, d) np.ndarray`
    Field number 0"""
Unique2DFloats.ids.__doc__ = """`(n, d) np.ndarray`
    Field number 1"""
Unique2DFloats.inverse.__doc__ = """`(n, d) np.ndarray`
    Field number 2"""
Unique2DFloats.intersection.__doc__ = """`(m) list of list`
  given original array's index, returns overlapping arrays, including itself.
  Field number 3
"""

Unique2DIntegers = namedtuple(
        "Unique2DIntegers", ["values", "ids", "inverse", "counts"]
)
"""namedtuple to hold unique information of integer type arrays.
Similar approach to Unique2DFloats.
"""

Unique2DIntegers.values.__doc__ = """`(n, d) np.ndarray`
    Field number 0"""
Unique2DIntegers.ids.__doc__ = """`(n) np.ndarray`
    Field number 1"""
Unique2DIntegers.inverse.__doc__ = """`(m) np.ndarray`
    Field number 2"""
Unique2DIntegers.counts.__doc__ = """`(n) np.ndarray`
    Field number 3"""


class ComputedMeshData(ComputedData):
    """A class to hold computed-mesh-data.

    Subclassed to keep its own dependency info.
    """
    pass



"""gustaf/spline/create.py.

Creates splines.
"""

import numpy as np
from splinepy.spline import RequiredProperties

from gustaf import utils
from gustaf import settings


def extruded(spline, extrusion_vector=None):
    """Extrudes Splines.

    Parameters
    ----------
    spline: GustafSpline
    extrusion_vector: np.ndarray
    """
    from gustaf.spline.base import GustafSpline

    # Check input type
    if not issubclass(type(spline), GustafSpline):
        raise NotImplementedError("Extrude only works for splines")

    # Check extrusion_vector
    if extrusion_vector is not None:
        # make flat extrusion_vector
        extrusion_vector = np.asarray(extrusion_vector).ravel()
    else:
        raise ValueError("No extrusion extrusion_vector given")

    # Check extrusion_vector dimension
    # formulate correct cps
    if spline.dim == extrusion_vector.shape[0]:
        cps = spline.control_points
    elif spline.dim < extrusion_vector.shape[0]:
        expansion_dimension = extrusion_vector.shape[0] - spline.dim
        # one smaller dim is allowed
        # warn that we assume new dim is all zero
        utils.log.warning(
                f"Given extrusion vector is {expansion_dimension} dimension "
                "bigger than spline's dim. Assuming 0.0 entries for "
                "new dimension.",
        )
        cps = np.hstack(
                (
                        spline.control_points,
                        np.zeros(
                                (
                                        len(spline.control_points),
                                        expansion_dimension
                                )
                        )
                )
        )
    else:
        raise ValueError(
                "Dimension Mismatch between extrusion extrusion vector "
                "and spline."
        )

    # Start Extrusion
    spline_dict = dict()

    spline_dict["degrees"] = np.concatenate((spline.degrees, [1]))
    spline_dict["control_points"] = np.vstack((cps, cps + extrusion_vector))
    if "knot_vectors" in RequiredProperties.of(spline):
        spline_dict["knot_vectors"] = spline.knot_vectors + [[0, 0, 1, 1]]
    if "weights" in RequiredProperties.of(spline):
        spline_dict["weights"] = np.concatenate(
                (spline.weights, spline.weights)
        )

    return type(spline)(**spline_dict)


def revolved(
        spline,
        axis=None,
        center=None,
        angle=None,
        n_knot_spans=None,
        degree=True
):
    """Revolve spline around an axis and extend its parametric dimension.

    Parameters
    ----------
    spline : GustafSpline
      Basis-Spline to be revolved
    axis : np.ndarray
      Axis of revolution
    center : np.ndarray
      Center of revolution
    angle : float
    n_knot_spans : int
      number of non-zero knot-elements for result-spline (if applicable)
    degree : bool
      use degrees instead of radiant

    Returns
    -------
    spline : GustafSpline
    """

    from gustaf.spline.base import GustafSpline

    # Check input type
    if not issubclass(type(spline), GustafSpline):
        raise NotImplementedError("Revolutions only works for splines")

    # Check axis
    if axis is not None:
        # Transform into numpy array
        axis = np.asarray(axis).ravel()
        # Check Axis dimension
        if (spline.control_points.shape[1] > axis.shape[0]):
            raise ValueError(
                    "Dimension Mismatch between extrusion axis and spline."
            )
        elif (spline.control_points.shape[1] < axis.shape[0]):
            utils.log.warning(
                    "Control Point dimension is smaller than axis dimension,"
                    " filling with zeros"
            )
            expansion_dimension = axis.shape[0] - spline.dim
            cps = np.hstack(
                    (
                            spline.control_points,
                            np.zeros(
                                    (
                                            len(spline.control_points),
                                            expansion_dimension
                                    )
                            )
                    )
            )
        else:
            cps = np.copy(spline.control_points)

        # Make sure axis is normalized

        axis_norm = np.linalg.norm(axis)
        if not np.isclose(axis_norm, 0, atol=settings.TOLERANCE):
            axis = axis / axis_norm
        else:
            raise ValueError("Axis-norm is too close to zero.")
    else:
        cps = np.copy(spline.control_points)
        if spline.control_points.shape[1] == 3:
            raise ValueError("No rotation axis given")

    # Set Problem dimension
    problem_dimension = cps.shape[1]

    # Make sure axis is ignored for 2D
    if problem_dimension == 2:
        axis = None

    # Update angle
    if degree:
        angle = np.radians(angle)

    # Init center
    if center is not None:
        center = np.asarray(center).ravel()
        # Check Axis dimension
        if not (problem_dimension == center.shape[0]):
            raise ValueError(
                    "Dimension Mismatch between axis and center of rotation."
            )
        cps -= center

    # The parametric dimension is independent of the revolution but the
    # rotation-matrix is only implemented for 2D and 3D problems
    if not (cps.shape[1] == 2 or cps.shape[1] == 3):
        raise NotImplementedError(
                "Sorry,"
                "revolutions only implemented for 2D and 3D splines"
        )

    # Angle must be (0, pi) non including
    # Rotation is always performed in half steps
    PI = np.pi
    minimum_n_knot_spans = int(
            np.ceil(np.abs((angle + settings.TOLERANCE) / PI))
    )
    if (n_knot_spans) is None or (n_knot_spans < minimum_n_knot_spans):
        n_knot_spans = minimum_n_knot_spans

    if "Bezier" in spline.whatami:
        if n_knot_spans > 1:
            raise ValueError(
                    "Revolutions are only supported for angles up to 180 "
                    "degrees for Bezier type splines as they consist of only "
                    "one knot span"
            )

    # Determine auxiliary values
    rot_a = angle / (2 * n_knot_spans)
    half_counter_angle = PI / 2 - rot_a
    weight = np.sin(half_counter_angle)
    factor = 1 / weight

    # Determine rotation matrix
    rotation_matrix = utils.arr.rotation_matrix_around_axis(
            axis=axis, rotation=rot_a, degree=False
    ).T

    # Start Extrusion
    spline_dict = dict()

    spline_dict["degrees"] = np.concatenate((spline.degrees, [2]))

    spline_dict["control_points"] = cps
    end_points = cps
    for i_segment in range(n_knot_spans):
        # Rotate around axis
        mid_points = np.matmul(end_points, rotation_matrix)
        end_points = np.matmul(mid_points, rotation_matrix)
        # Move away from axis using dot-product tricks
        if problem_dimension == 3:
            mp_scale = axis * np.dot(mid_points, axis).reshape(-1, 1)
            mid_points = (mid_points - mp_scale) * factor + mp_scale
        else:
            mid_points *= factor
        spline_dict["control_points"] = np.concatenate(
                (spline_dict["control_points"], mid_points, end_points)
        )

    if "knot_vectors" in RequiredProperties.of(spline):
        kv = [0, 0, 0]
        [kv.extend([i + 1, i + 1]) for i in range(n_knot_spans - 1)]
        spline_dict["knot_vectors"] = spline.knot_vectors + [
                kv + [n_knot_spans + 1] * 3
        ]
    if "weights" in RequiredProperties.of(spline):
        mid_weights = spline.weights * weight
        spline_dict["weights"] = spline.weights
        for i_segment in range(n_knot_spans):
            spline_dict["weights"] = np.concatenate(
                    (spline_dict["weights"], mid_weights, spline.weights)
            )
    else:
        utils.log.warning(
                "True revolutions are only possible for rational spline types."
                "\nCreating Approximation."
        )

    if center is not None:
        spline_dict["control_points"] += center

    return type(spline)(**spline_dict)


def line(points):
    """Create a spline with the provided points as control points.

    Parameters
    ----------
    points: (n, d) numpy.ndarray
      npoints x ndims array of control points

    Returns
    -------
    line: BSpline
      Spline degree [1].
    """
    from gustaf import BSpline

    # lines have degree 1
    degree = 1

    cps = np.array(points)
    nknots = cps.shape[0] + degree + 1

    knots = np.concatenate(
            (
                    np.full(degree, 0.),
                    np.linspace(0., 1., nknots - 2 * degree),
                    np.full(degree, 1.0),
            )
    )

    spline = BSpline(
            control_points=cps, knot_vectors=[knots], degrees=[degree]
    )

    return spline


def arc(radius=1., angle=90., n_knot_spans=None, start_angle=0., degree=True):
    """Creates a 1-D arc as Rational Bezier or NURBS with given radius and
    angle. The arc lies in the x-y plane and rotates around the z-axis.

    Parameters
    ----------
    radius : float, optional
      radius of the arc, defaults to 1
    angle : float, optional
      angle of the section of the arc, defaults to 90 degrees
    n_knot_spans : int
      Number of knot spans, by default minimum number for angle is used.
    start_angle : float, optional
      starting point of the angle, by default 0.
    degree: bool, optional
      degrees for angle used, by default True

    Returns
    -------
    arc: NURBS or RationalBezier
    """
    from gustaf import RationalBezier

    # Define point spline of degree 0 at starting point of the arc
    if degree:
        start_angle = np.radians(start_angle)
        angle = np.radians(angle)
    start_point = [radius * np.cos(start_angle), radius * np.sin(start_angle)]
    point_spline = RationalBezier(
            degrees=[0], control_points=[start_point], weights=[1.]
    )
    # Bezier splines only support angles lower than 180 degrees
    if abs(angle) >= np.pi or n_knot_spans > 1:
        point_spline = point_spline.nurbs

    # Revolve
    arc_attribs = point_spline.create.revolved(
            angle=angle, n_knot_spans=n_knot_spans, degree=degree
    ).todict()
    # Remove the first parametric dimenions, which is only a point and
    # only used for the revolution
    arc_attribs['degrees'] = list(arc_attribs['degrees'])[1:]
    if "knot_vectors" in RequiredProperties.of(point_spline):
        arc_attribs['knot_vectors'] = list(arc_attribs['knot_vectors'])[1:]

    return type(point_spline)(**arc_attribs)


def circle(radius=1., n_knot_spans=3):
    """Circle (parametric dim = 1) with radius r in the x-y plane around the
    origin. The spline has an open knot vector and degree 2.

    Parameters
    ----------
    radius : float, optional
      radius, defaults to one
    n_knots_spans : int, optional
      number of knot spans, defaults to 3

    Returns
    -------
    circle: NURBS
    """
    return arc(radius=radius, angle=360, n_knot_spans=n_knot_spans)


def box(*lengths):
    """ND box (hyperrectangle).

    Parameters
    ----------
    *lengths: list(float)

    Returns
    -------
    ndbox: Bezier
    """
    from gustaf import Bezier

    # may dim check here?

    # starting point
    ndbox = Bezier(degrees=[1], control_points=[[0], [lengths[0]]])
    # use extrude
    for i, l in enumerate(lengths[1:]):
        ndbox = ndbox.create.extruded([0] * int(i + 1) + [l])

    return ndbox


def plate(radius=1.):
    """Creates a biquadratic 2-D spline in the shape of a plate with given
    radius.

    Parameters
    ----------
    radius : float, optional
      Radius of the plate, defaults to one

    Returns
    -------
    plate: RationalBezier
    """
    from gustaf.spline import RationalBezier

    degrees = [2, 2]
    control_points = np.array(
            [
                    [-0.5, -0.5],
                    [0., -1.],
                    [0.5, -0.5],
                    [-1., 0.],
                    [0., 0.],
                    [1., 0.],
                    [-0.5, 0.5],
                    [0., 1.],
                    [0.5, 0.5],
            ]
    ) * radius
    weights = np.tile([1., 1 / np.sqrt(2)], 5)[:-1]

    return RationalBezier(
            degrees=degrees, control_points=control_points, weights=weights
    )


def disk(
        outer_radius,
        inner_radius=None,
        angle=360.,
        n_knot_spans=None,
        degree=True
):
    """Surface spline describing a potentially hollow disk with quadratic
    degree along curved dimension and linear along thickness. The angle
    describes the returned part of the disk.

    Parameters
    ----------
    outer_radius : float
      Outer radius of the disk
    inner_radius : float, optional
      Inner radius of the disk, in case of hollow disk, by default 0.
    angle : float, optional
      Rotational angle, by default 360. describing a complete revolution
    n_knot_spans : int, optional
      Number of knot spans, by default 4

    Returns
    -------
    disk: NURBS
      Surface NURBS of degrees (1,2)
    """

    from gustaf.spline import NURBS

    if inner_radius is None:
        inner_radius = 0.

    cps = np.array([[inner_radius, 0.], [outer_radius, 0.]])
    weights = np.ones([cps.shape[0]])
    knots = np.repeat([0., 1.], 2)

    return NURBS(
            control_points=cps,
            knot_vectors=[knots],
            degrees=[1],
            weights=weights
    ).create.revolved(angle=angle, n_knot_spans=n_knot_spans, degree=degree)


def torus(
        torus_radius,
        section_outer_radius,
        section_inner_radius=None,
        torus_angle=None,
        section_angle=None,
        section_n_knot_spans=4,
        torus_n_knot_spans=4,
        degree=True,
):
    """Creates a volumetric NURBS spline describing a torus revolved around the
    x-axis. Possible cross-sections are plate, disk (yielding a tube) and
    section of a disk.

    Parameters
    ----------
    torus_radius : float
      Radius of the torus
    section_outer_radius : float
      Radius of the section of the torus
    section_inner_radius : float, optional
      Inner radius in case of hollow torus, by default 0.
    torus_angle : float, optional
      Rotational anlge of the torus, by default None, giving a complete
      revolution
    section_angle : float, optional
      Rotational angle, by default None, yielding a complete revolution
    section_n_knot_spans : float, optional
      Number of knot spans along the cross-section, by default 4
    torus_n_knot_spans : float, optional
      Number of knot spans along the torus, by default 4

    Returns
    -------
    torus: NURBS
      Volumetric spline in the shape of a torus with degrees (1,2,2)
    """

    if torus_angle is None:
        torus_angle = 2 * np.pi
        degree = False

    if section_angle is None:
        section_angle = 2 * np.pi
        section_angle_flag = False
        degree = False
    else:
        section_angle_flag = True

    if section_inner_radius is None:
        section_inner_radius = 0
        section_inner_radius_flag = False
    else:
        section_inner_radius_flag = True

    # Create the cross-section
    if not section_angle_flag and not section_inner_radius_flag:
        cross_section = plate(section_outer_radius)
        # For more than 180 degree only NURBS can be used
        if abs(torus_angle) >= np.pi:
            cross_section = cross_section.nurbs
    else:
        cross_section = disk(
                outer_radius=section_outer_radius,
                inner_radius=section_inner_radius,
                n_knot_spans=section_n_knot_spans,
                angle=section_angle,
                degree=degree
        )

    # Create a surface spline representing a disk and move it from the origin
    cross_section.control_points[:, 1] += torus_radius

    return cross_section.create.revolved(
            axis=[1., 0, 0],
            center=np.zeros(3),
            angle=torus_angle,
            n_knot_spans=torus_n_knot_spans,
            degree=degree
    )


def sphere(
        outer_radius,
        inner_radius=None,
        angle=360.,
        n_knot_spans=None,
        degree=True
):
    """Creates a volumetric spline describing a sphere with radius R.

    Parameters
    ----------
    outer_radius : float
      Outer radius of the sphere
    inner_radius : float, optional
      Inner radius of the potentially hollow sphere.
    angle : float
      Rotational angle around x-axis, by default each 360
      (describing a complete revolution)
    n_knot_spans : int
      Number of knot spans

    Returns
    -------
    sphere: NURBS
      Volumetric NURBS with degrees (1,2,2)
    """

    if inner_radius is None:
        sphere = plate(outer_radius).nurbs.create.revolved(
                axis=[1, 0, 0],
                center=[0, 0, 0],
                angle=angle,
                n_knot_spans=n_knot_spans,
                degree=degree
        )
    else:
        inner_radius = float(inner_radius)
        sphere = disk(outer_radius, inner_radius).nurbs.create.revolved(
                angle=angle, n_knot_spans=n_knot_spans, degree=degree
        )
    return sphere


def cone(
        outer_radius,
        height,
        inner_radius=None,
        volumetric=True,
        angle=360.,
        degree=True
):
    """Creates a cone with circular base.

    Parameters
    ----------
    radius : float
      Radius of the base
    height : float
      Height of the cone
    volumetric : bool, optional
      Parameter whether surface or volume spline, by default True
    angle : float
      Rotation angle in degrees, only used for solid model

    Returns
    -------
    cone: NURBS
      Volumetric or surface NURBS descibing a cone
    """

    if volumetric:
        ground = disk(
                outer_radius,
                inner_radius=inner_radius,
                angle=angle,
                degree=degree
        )
    else:
        ground = circle(outer_radius)

    # Extrude in z
    cone = ground.create.extruded([0, 0, height])
    # Move all upper control points to one
    cone.control_points[np.isclose(cone.control_points[:, -1],
                                   height)] = [0, 0, height]

    return cone


def pyramid(width, length, height):
    """Creates a volumetric spline in the shape of a pyramid with linear degree
    in every direction.

    Parameters
    ----------
    width : float
      Dimension of base in x-axis
    lenght : float
      Dimension of base in y-axis
    height : float
      Height in z-direction

    Returns
    -------
    pyramid: BSpline
      Volumetric linear spline in the shape of a pyramid
    """

    # Create box
    p = box(width, length, height)

    # Collapse all upper points on one control point
    p.control_points[4:, :] = [width / 2, length / 2, height]

    return p


class Creator:
    """Helper class to build new splines from existing geometries.

    Examples
    ---------
    >>> myspline = <your-spline>
    >>> spline_faces = myspline.create.extrude(vector=[3,1,3])
    """

    def __init__(self, spl):
        self.spline = spl

    def extruded(self, *args, **kwargs):
        return extruded(self.spline, *args, **kwargs)

    def revolved(self, *args, **kwargs):
        return revolved(self.spline, *args, **kwargs)



from gustaf.utils import arr
from gustaf.utils import connec
from gustaf.utils import log
from gustaf.utils import tictoc

# Alias
connectivity = connec

__all__ = [
        "arr",
        "connec",
        "connectivity",
        "log",
        "tictoc",
]



"""gustaf/spline/extract.py.

Extract operations. Both discrete and spline extraction.
"""

import itertools

import numpy as np

from gustaf import utils
from gustaf.vertices import Vertices
from gustaf.edges import Edges
from gustaf.faces import Faces
from gustaf.volumes import Volumes

from gustaf.spline._utils import to_res_list


def edges(
        spline,
        resolution=100,
        extract_dim=None,
        extract_knot=None,
        all_knots=False,
):
    """Extract edges (lines) from a given spline. Only entity you can extract
    without dimension limit.

    Parameters
    -----------
    spline: Spline
    resolution: int
    extract_dim: int
      Parametric dimension to extract.
    extract_knot: list
      (spline.para_dim - 1,) shaped knot location along extract_dim
    all_knots: bool
      Switch to allow all knot-line extraction.

    Returns
    --------
    edges: Edges
    """
    if not all_knots:
        resolution = int(resolution)

    if spline.para_dim == 1:
        return Edges(
                vertices=spline.sample(resolution),
                edges=utils.connec.range_to_edges(
                        (0, resolution),
                        closed=False,
                )
        )

    else:
        # This should be possible for spline of any dimension.
        # As long as it satisfies the following condition
        if extract_knot is not None:
            if len(extract_knot) != spline.para_dim - 1:
                raise ValueError(
                        "Must satisfy len(extract_knot) == spline.para_dim -1."
                )

        # This may take awhile.
        if all_knots:
            edgess = []  # edges' is not a valid syntax
            unique_knots = np.array(spline.unique_knots, dtype=object)
            for i in range(spline.para_dim):
                mask = np.ones(spline.para_dim, dtype=bool)
                mask[i] = False
                # gather knots along current knot
                extract_knot_queries = list(
                        itertools.product(*unique_knots[mask])
                )

                for ekq in extract_knot_queries:
                    edgess.append(edges(spline, resolution[i], i, ekq, False))

            return Edges.concat(edgess)

        # Get parametric points to extract
        queries = np.empty(
                (resolution, spline.para_dim),
                dtype="float64",  # hardcoded for splinelibpy
                order="C",  # hardcoded for splinelibpy
        )
        # get ~extract_dim
        not_ed = np.arange(spline.para_dim).tolist()
        not_ed.pop(extract_dim)
        queries[:, not_ed] = extract_knot

        # get knot extrema
        uniq_knots = spline.unique_knots[extract_dim]
        min_knot_position = min(uniq_knots)
        max_knot_position = max(uniq_knots)

        queries[:, extract_dim] = np.linspace(
                min_knot_position,
                max_knot_position,
                resolution,
        )

        return Edges(
                vertices=spline.evaluate(queries),
                edges=utils.connec.range_to_edges(
                        (0, resolution),
                        closed=False,
                )
        )


def faces(
        spline,
        resolutions,
):
    """Extract faces from spline. Valid iff para_dim is one of the followings:
    {2, 3}. In case of {3}, it will return only surfaces. If internal faces are
    desired, used `spline.extract.volumes().faces()`. Note that dimension
    higher than 3 is not showable.

    Parameters
    -----------
    spline: BSpline or NURBS
    resolutions: int or list

    Returns
    --------
    faces: faces
    """
    resolutions = to_res_list(resolutions, spline.para_dim)

    if spline.para_dim == 2:
        return Faces(
                vertices=spline.sample(resolutions),
                faces=utils.connec.make_quad_faces(resolutions),
        )

    elif spline.para_dim == 3:
        # TODO: use spline extraction routine to first extract
        # spline, extract faces, merge vertices.

        # Spline to surfaces
        vertices = []
        faces = []
        offset = 0
        # accomodate bezier Splines
        ukvs = spline.unique_knots

        for i in range(spline.para_dim):
            extract = i
            # Get extracting dimension
            extract_along = [0, 1, 2]
            extract_along.pop(extract)

            # Extract range
            extract_range = [
                    [
                            min(ukvs[extract_along[0]]),
                            max(ukvs[extract_along[0]]),
                    ],
                    [
                            min(ukvs[extract_along[1]]),
                            max(ukvs[extract_along[1]]),
                    ],
            ]

            extract_list = [
                    min(ukvs[extract]),
                    max(ukvs[extract]),
            ]

            # surface point queries (spq)
            spq = np.linspace(
                    extract_range[0][0],
                    extract_range[0][1],
                    resolutions[extract_along[0]],
            ).reshape(-1, 1)

            # expand horizontally and init with 1
            spq = np.hstack((spq, np.ones((len(spq), 1))))
            spq = np.vstack(
                    np.linspace(
                            spq * [1, extract_range[1][0]],
                            spq * [1, extract_range[1][1]],
                            resolutions[extract_along[1]],
                    )
            )

            # expand horizontally and init with 1
            spq = np.hstack((spq, np.ones((len(spq), 1))))
            spq = np.vstack(
                    np.linspace(
                            spq * [1, 1, extract_list[0]],
                            spq * [1, 1, extract_list[1]], 2
                    )
            )

            surface_point_queries = utils.arr.make_c_contiguous(
                    spq,
                    dtype="float64",
            )
            sorted_ids = np.argsort(
                    [extract_along[0], extract_along[1], extract]
            )
            surface_point_queries = surface_point_queries[:, sorted_ids]

            vertices.append(
                    spline.evaluate(
                            surface_point_queries[:int(
                                    surface_point_queries.shape[0] / 2
                            )]
                    )
            )

            if len(faces) != 0:
                offset = faces[-1].max() + 1

            tmp_faces = utils.connec.make_quad_faces(
                    [
                            resolutions[extract_along[0]],
                            resolutions[extract_along[1]],
                    ]
            )

            faces.append(tmp_faces + int(offset))

            vertices.append(
                    spline.evaluate(
                            surface_point_queries[
                                    int(surface_point_queries.shape[0] / 2):]
                    )
            )

            offset = faces[-1].max() + 1

            faces.append(tmp_faces + int(offset))

        # make faces and merge vertices before returning
        f = Faces(vertices=np.vstack(vertices), faces=np.vstack(faces))
        f.merge_vertices()

        return f

    else:
        raise ValueError("Invalid spline to make faces.")


def volumes(spline, resolutions):
    """Extract volumes from spline. Valid iff spline.para_dim == 3.

    Parameters
    -----------
    spline: BSpline or NURBS
    resolutions:

    Returns
    --------
    volumes: Volumes
    """
    if spline.para_dim != 3:
        raise ValueError(
                "Volume extraction from a spline is only valid for "
                "para_dim: 3 dim: 3 splines."
        )

    return Volumes(
            vertices=spline.sample(resolutions),
            volumes=utils.connec.make_hexa_volumes(resolutions),
    )


def control_points(spline):
    """Extracts control points and return as vertices. Same can be achieved by
    doing `gustaf.Vertices(spline.control_points)`

    Parameters
    -----------
    spline: BSpline or NURBS

    Returns
    --------
    cps_as_Vertices: Vertices
    """
    return Vertices(spline.control_points)


def control_edges(spline):
    """Extract control edges (mesh). Valid iff para_dim is 1.

    Parameters
    -----------
    edges: BSpline or NURBS

    Returns
    --------
    edges: Edges
    """
    if spline.para_dim != 1:
        raise ValueError("Invalid spline type!")

    return Edges(
            vertices=spline.control_points,
            edges=utils.connec.range_to_edges(
                    len(spline.control_points), closed=False
            )
    )


def control_faces(spline):
    """Extract control face (mesh). Valid iff para_dim is 2.

    Parameters
    -----------
    spline: BSpline or NURBS

    Returns
    --------
    faces: Faces
    """
    if spline.para_dim != 2:
        raise ValueError("Invalid spline type!")

    return Faces(
            vertices=spline.control_points,
            faces=utils.connec.make_quad_faces(
                    spline.control_mesh_resolutions
            ),
    )


def control_volumes(spline):
    """Extract control volumes (mesh). Valid iff para_dim is 3.

    Parameters
    -----------
    spline: BSpline or NURBS

    Returns
    --------
    volumes: Volumes
    """
    if spline.para_dim != 3:
        raise ValueError("Invalid spline type!")

    return Volumes(
            vertices=spline.control_points,
            volumes=utils.connec.make_hexa_volumes(
                    spline.control_mesh_resolutions
            ),
    )


def control_mesh(spline, ):
    """Calls control_edges, control_faces, control_volumes based on current
    spline.

    Parameters
    -----------
    None

    Returns
    --------
    control_mesh: Edges or Faces or Volumes
    """
    if spline.para_dim == 1:
        return control_edges(spline)
    elif spline.para_dim == 2:
        return control_faces(spline)
    elif spline.para_dim == 3:
        return control_volumes(spline)
    else:
        raise ValueError(
                "Invalid para_dim to extract control_mesh. "
                "Supports 1 to 3."
        )


def beziers(spline):
    """Extracts Bezier-type objects of any spline-type object.

  Parameters
  ----------
  spline : Gustaf-Spline

  Returns
  -------
  bezier_list : list<bezier-types>
  """
    from gustaf.spline.base import Bezier, RationalBezier

    if "Bezier" in spline.whatami:
        return [spline]
    elif "BSpline" in spline.whatami:
        return [
                Bezier(**s.todict())
                for s in super(type(spline), spline).extract_bezier_patches()
        ]
    elif "NURBS" in spline.whatami:
        return [
                RationalBezier(**s.todict())
                for s in super(type(spline), spline).extract_bezier_patches()
        ]
    else:
        raise TypeError("Unknown Spline-Type.")


def spline(spline, para_dim, split_plane):
    """Extract a subspline from a given representation.

    Parameters
    ----------
    para_dim : int
      parametric dimension to be extrac ted
    split_plane : float / tuple<float, float>
      intervall or value in parametric space to be extracted from the spline
      representation

    Returns
    -------
    spline
    """
    from gustaf.spline.base import GustafSpline

    # Check type
    if not issubclass(type(spline), GustafSpline):
        raise TypeError("Unknown spline representation passed to subspline")

    # Check arguments for sanity
    if para_dim > spline.para_dim:
        raise ValueError(
                "Requested parametric dimension exceeds spline's parametric"
                " dimensionality."
        )
    if isinstance(split_plane, list):
        if not (
                (len(split_plane) == 2)
                and (isinstance(split_plane[0], float))
                and (isinstance(split_plane[0], float))
        ):
            raise ValueError(
                    "Range must be float or tuple of floats with length 2"
            )
    elif not isinstance(split_plane, float):
        raise ValueError(
                "Range must be float or tuple of floats with length 2"
        )
    else:
        # Convert float to tuple to facilitate
        split_plane = list([split_plane])

    # Check if is bezier-type
    is_bezier = "Bezier" in spline.whatami
    is_rational = "weights" in spline.required_properties
    if is_bezier:
        if is_rational:
            spline_copy = spline.nurbs
        else:
            spline_copy = spline.bspline
    else:
        spline_copy = spline.copy()

    for _ in range(spline_copy.degrees[para_dim]):
        # Will do nothing if spline already has sufficient number of knots
        # at given position
        spline_copy.insert_knots(para_dim, split_plane)

    # Start extraction
    cps_res = spline_copy.control_mesh_resolutions
    start_id = spline_copy.knot_vectors[para_dim].index(split_plane[0])
    end_id = spline_copy.knot_vectors[para_dim].index(split_plane[-1])
    para_dim_ids = np.arange(np.prod(cps_res))
    for i_pd in range(para_dim):
        para_dim_ids -= para_dim_ids % cps_res[i_pd]
        para_dim_ids = para_dim_ids // cps_res[i_pd]
    # indices are shifted by one
    para_dim_ids = para_dim_ids % cps_res[para_dim] + 1

    # Return new_spline
    spline_info = {}
    spline_info["control_points"] = spline_copy.cps[(para_dim_ids >= start_id)
                                                    & (para_dim_ids <= end_id)]
    spline_info["degrees"] = spline_copy.degrees.tolist()
    if start_id == end_id:
        spline_info["degrees"].pop(para_dim)
    if not is_bezier:
        spline_info["knot_vectors"] = spline_copy.knot_vectors.copy()
        if start_id == end_id:
            spline_info["knot_vectors"].pop(para_dim)
        else:
            spline_info["knot_vectors"][para_dim] = (
                    [spline_copy.knot_vectors[para_dim][start_id]]
                    + spline_copy.knot_vectors[para_dim]
                    [start_id:(end_id + spline_copy.degrees[para_dim])] + [
                            spline_copy.knot_vectors[para_dim]
                            [(end_id + spline_copy.degrees[para_dim] - 1)]
                    ]
            )
    if is_rational:
        spline_info["weights"] = spline_copy.weights[
                (para_dim_ids >= start_id) & (para_dim_ids <= end_id)]

    return type(spline)(**spline_info)


class Extractor:
    """Helper class to allow direct extraction from spline obj (BSpline or
    NURBS). Internal use only.

    Examples
    ---------
    >>> myspline = <your-spline>
    >>> spline_faces = myspline.extract.faces()
    """

    def __init__(self, spl):
        self._spline = spl

    def edges(self, *args, **kwargs):
        return edges(self._spline, *args, **kwargs)

    def faces(self, *args, **kwargs):
        return faces(self._spline, *args, **kwargs)

    def volumes(self, *args, **kwargs):
        return volumes(self._spline, *args, **kwargs)

    def control_points(self):
        return control_points(self._spline)

    def control_edges(self):
        return control_edges(self._spline)

    def control_faces(self):
        return control_faces(self._spline)

    def control_volumes(self):
        return control_volumes(self._spline)

    def control_mesh(self):
        return control_mesh(self._spline)

    def beziers(self):
        return beziers(self._spline)

    def spline(self, splittin_plane=None, interval=None):
        """Extract a spline from a spline.

        Use a (number of) splitting planes to extract a subsection from the
        parametric domain of it.

        Parameters
        ----------
        splitting_plane : int / dictionary (int : (float))
          if integer : parametric dimension to be extracted
          if dictionary : list of splitting planes and ranges to be passed
        interval : float / tuple<float,float>
          intervall or value in parametric space to be extracted from the
          spline representation
        Returns
        -------
        spline
        """
        if isinstance(splittin_plane, dict):
            if interval is not None:
                raise ValueError("Arguments incompatible expect dictionary")
            splittin_plane = dict(
                    sorted(splittin_plane.items(), key=lambda x: x[0])[::-1]
            )
            spline_copy = self._spline.copy()
            for key, item in splittin_plane.items():
                spline_copy = spline(spline_copy, key, item)
            return spline_copy
        else:
            return spline(self._spline, splittin_plane, interval)



"""gustaf/spline/base.py.

Base for splines. Contains show and inherited classes from `spline`.
"""

import abc
import logging

import splinepy
import numpy as np

from gustaf import settings
from gustaf import show as showmodule
from gustaf._base import GustafBase
from gustaf.vertices import Vertices
from gustaf.spline.create import Creator
from gustaf.spline.extract import Extractor
from gustaf.spline.proximity import Proximity
from gustaf.spline._utils import to_res_list


def show(
        spline,
        resolutions=100,
        control_points=True,
        knots=True,
        show_fitting_queries=True,
        return_discrete=False,
        return_showable=False,
        backend=None,
        # From here, | only relevant if "vedo" is backend.
        #            V
        parametric_space=False,
        color=None,
        surface_alpha=1,
        control_points_alpha=.8,
        lighting="glossy",
        control_point_ids=True,
        color_spline=None,
        cmap=None,  # only required
        **kwargs,
):
    """Shows splines with various options. They are excessively listed, so that
    it can be adjustable.

    Parameters
    -----------
    spline: BSpline or NURBS
    resolutions: int or (spline.para_dim,) array-like
    control_points: bool
    knots: bool
    show_fitting_queries: bool
    return_discrete: bool
      Return dict of gustaf discrete objects, for example,
      {Vertices, Edges, Faces}, instead of opening a window
    return_showable: bool
      Return dict of showable objects.
    parametric_space: bool
      Only relevant for `vedo` backend.
    color: str
      Default is None. Black for curves, else green.
    surface_alpha: float
      Only relevant for `vedo` backend. Effective range [0, 1].
    lighting: str
      Only relevant for `vedo` backend.
    control_point_ids: bool
    color_spline: str
    cmap: str
      Only relevant for `vedo` backend and color_spline is not None.

    Returns
    --------
    things_to_show: dict
      iff return_discrete==True, dict of gustaf objects that are showable.
      iff return_showable==True, dict of backend objects that are showable.
    """
    # Showing is only possible for following splines
    allowed_dim_combo = (
            (1, 2),
            (1, 3),
            (2, 2),
            (2, 3),
            (3, 3),
    )
    if (spline.para_dim, spline.dim) not in allowed_dim_combo:
        raise ValueError("Sorry, can't show given spline.")

    # determine backend
    if backend is None:
        backend = settings.VISUALIZATION_BACKEND

    # Prepare correct form of resolution input
    resolutions = to_res_list(resolutions, spline.para_dim)

    # Prepare things to show dict.
    things_to_show = dict()

    # (discretized) spline itself with basic color scheme.
    if spline.para_dim == 1:
        sp = spline.extract.edges(resolutions[0])
        if color is None:
            color = "black"
        sp.vis_dict.update(c=color, lw=8)

    if spline.para_dim == 2 or spline.para_dim == 3:
        sp = spline.extract.faces(resolutions)
        if color is None:
            color = "green"
        sp.vis_dict.update(c=color)
        # If

    things_to_show.update(spline=sp)

    # control_points = control mesh + control_points
    # control mesh always takes unique edge representation.
    if control_points:
        control_mesh = spline.extract.control_mesh()
        if spline.para_dim != 1:
            control_mesh = control_mesh.toedges(unique=True)

        # Set alpha to < 1, so that they don't "overshadow" spline
        control_mesh.vis_dict.update(c="red", lw=4, alpha=control_points_alpha)
        things_to_show.update(control_mesh=control_mesh)  # mesh itself
        # Add big vertices to emphasize cps.
        cps = control_mesh.tovertices()
        cps.vis_dict.update(c="red", r=10, alpha=control_points_alpha)
        things_to_show.update(control_points=cps)  # only points

    if knots:
        # Knot lines for non-curve splines.
        # Knot for curves are only added for vedo backend.
        if spline.para_dim > 1:
            knot_lines = spline.extract.edges(resolutions, all_knots=True)
            knot_lines.vis_dict.update(c="black", lw=3)
            things_to_show.update(knots=knot_lines)

    if show_fitting_queries and hasattr(spline, "_fitting_queries"):
        fitting_queries = Vertices(spline._fitting_queries)
        fitting_queries.vis_dict.update(c="blue", r=10)
        things_to_show.update(fitting_queries=fitting_queries)

    # Return here, if backend is not vedo
    if not backend.startswith("vedo"):
        # turn everything into backend showables
        if return_showable:
            for key, gusobj in things_to_show.items():
                things_to_show.update({key: gusobj.showable(backend=backend)})

            return things_to_show

        elif return_discrete:
            return things_to_show

        else:
            showmodule.show(list(things_to_show.values()))
            return None

    # iff backend is vedo, we provide fancier visualization
    elif backend.startswith("vedo"):
        # return if showable is not desired
        # -> From now on we will directly work on vedo objects.
        if return_discrete and not return_showable:
            return things_to_show

        # turn all gus objects into gus objects.
        vedo_things = dict()
        for key, gusobj in things_to_show.items():
            vedo_things.update({key: gusobj.showable(backend=backend)})

        # apply lighting
        if lighting is not None:
            vedo_things["spline"].lighting(lighting)

        # adjust surface alpha
        if spline.para_dim > 1:
            vedo_things["spline"].alpha(surface_alpha)

        # add red points for at control points
        if control_points and control_point_ids:
            vedo_things.update(
                    control_point_ids=vedo_things["control_points"].
                    labels("id")
            )

        # add knots as "x" for curves
        if knots and spline.para_dim == 1:
            uks = spline.unique_knots[0]
            phys_uks = showmodule.make_showable(
                    Vertices(spline.evaluate([[uk] for uk in uks])),
                    backend=backend,
            )
            xs = ["x"] * len(uks)

            vedo_things.update(
                    knots=phys_uks.labels(xs, justify="center", c="green")
            )

        # generate parametric view of spline
        if parametric_space and spline.para_dim > 1:
            from vedo.addons import Axes
            from gustaf.create.spline import parametric_view
            from gustaf.utils.arr import bounds

            para_spline = parametric_view(spline)
            para_showables = show(
                    para_spline,
                    control_points=False,
                    return_showable=True,
                    lighting=lighting,
                    knots=knots,
                    parametric_space=False,
                    backend=backend,
            )
            # Make lines a bit thicker
            if knots:
                para_showables["knots"].lw(6)

            # Trick to show begin/end value
            bs = np.asarray(bounds(para_showables["spline"].points()))
            bs_diff_001 = (bs[1] - bs[0]) * 0.001
            lowerb = bs[0] - bs_diff_001
            upperb = bs[1] + bs_diff_001

            axes_config = dict(
                    xtitle="u",
                    ytitle="v",
                    xrange=[lowerb[0], upperb[0]],
                    yrange=[lowerb[1], upperb[1]],
                    tipSize=0,
                    xMinorTicks=3,
                    yMinorTicks=3,
                    xyGrid=False,
                    yzGrid=False,
            )

            if spline.para_dim == 3:
                axes_config.update(ztitle="w")
                axes_config.update(zrange=[lowerb[2], upperb[2]])
                axes_config.update(zMinorTicks=3)
                axes_config.update(zxGrid=False)

            para_showables.update(
                    axes=Axes(para_showables["spline"], **axes_config)
            )

        # showable return
        if return_showable:
            if parametric_space:
                vedo_things.update(parametric_spline=para_showables)
            return vedo_things

        # now, show
        if parametric_space:
            para_showables.update(description="Parametric View")
            vedo_things.update(description="Physical View")
            plt = showmodule.show_vedo(para_showables, vedo_things, **kwargs)

        else:
            plt = showmodule.show_vedo(vedo_things, **kwargs)

        return plt


class GustafSpline(GustafBase):

    @abc.abstractmethod
    def __init__(self):
        """Contructor as abstractmethod.

        This needs to be inherited first to make sure duplicating
        functions properly override splinepy.Spline
        """
        pass

    @property
    def extract(self):
        """Returns spline extracter. Can directly perform extractions available
        at `gustaf/spline/extract.py`. For more info, take a look at
        `gustaf/spline/extract.py`: Extracter.

        Examples
        ---------
        >>> splinefaces = spline.extract.faces()

        Parameters
        -----------
        None

        Returns
        --------
        spline_extracter: Extracter
        """
        return self._extractor

    @property
    def create(self):
        """Returns spline creator Can be used to create new splines using
        geometric relations.

        Examples
        --------
        >>> prism = spline.create.extrude(axis=[1,4,1])

        Parameters
        ----------
        None

        Returns
        spline.Creator
        """
        return self._creator

    @property
    def proximity(self):
        """Returns spline proximity helper. Can directly perform proximity
        queries available at `gustaf/spline/proximity.py`. For more info, take
        a look at `gustaf/spline/proximity.py`: Proximity.

        Examples
        ---------
        >>> closest_cp_ids = spline.proximity.closest_control_points(queries)
        >>> closest_cp_ids, distances =\
        ...    spline.proximity.closest_control_points(
        ...       queries,
        ...       return_distances=True
        ...    )

        Parameters
        -----------
        None

        Returns
        --------
        spline_proximity: Proximity
        """
        return self._proximity

    def show(self, **kwargs):
        """Equivalent to `gustaf.spline.base.show(**kwrags)`"""
        return show(self, **kwargs)

    def showable(self, **kwargs):
        """Equivalent to
        `gustaf.spline.base.show(return_showable=True,**kwargs)`"""
        return show(self, return_showable=True, **kwargs)

    def copy(self):
        """"""
        return type(self)(**self.todict())


class Bezier(GustafSpline, splinepy.Bezier):

    def __init__(
            self,
            degrees=None,
            control_points=None,
            spline=None,
    ):
        """Bezier of gustaf. Inherited from splinepy.Bezier and GustafSpline.

        Attributes
        -----------
        extract: Extractor
        create: Creator
        proximity: Proximity

        Parameters
        -----------
        degrees: (para_dim,) list-like
        control_points: (m, dim) list-like

        Returns
        --------
        None
        """
        super(splinepy.Bezier, self).__init__(
                degrees=degrees, control_points=control_points, spline=spline
        )

        self._extractor = Extractor(self)
        self._proximity = Proximity(self)
        self._creator = Creator(self)

    @property
    def bezier(self):
        """Returns same parametric representation as Bezier Spline.

        Parameters
        ----------
        None

        Returns
        -------
        same : Bezier
        """
        return self.copy()

    @property
    def rationalbezier(self):
        """Returns same parametric representation as Rational Bezier Spline.

        Parameters
        ----------
        None

        Returns
        -------
        same : RationalBezier
        """
        return RationalBezier(
                degrees=self.degrees,
                control_points=self.control_points,
                weights=np.ones(self.control_points.shape[0])
        )

    @property
    def bspline(self):
        """Returns same parametric representation as BSpline.

        Parameters
        -----------
        None

        Returns
        --------
        same_bspline : BSpline
        """
        return BSpline(
                degrees=self.degrees,
                control_points=self.control_points,
                knot_vectors=[
                        [0] * (self.degrees[i] + 1) + [1]
                        * (self.degrees[i] + 1) for i in range(self.para_dim)
                ]
        )

    @property
    def nurbs(self):
        """Returns same parametric representation as nurbs.

        Parameters
        -----------
        None

        Returns
        --------
        same_nurbs: NURBS
        """
        return self.bspline.nurbs


class RationalBezier(GustafSpline, splinepy.RationalBezier):

    def __init__(
            self,
            degrees=None,
            control_points=None,
            weights=None,
            spline=None,
    ):
        """Rational Bezier of gustaf. Inherited from splinepy.RationalBezier
        and GustafSpline.

        Attributes
        -----------
        extract: Extractor
        create: Creator
        proximity: Proximity

        Parameters
        -----------
        degrees: (para_dim,) list-like
        control_points: (m, dim) list-like
        weights : (m) list-like

        Returns
        --------
        None
        """
        super(splinepy.RationalBezier, self).__init__(
                degrees=degrees,
                control_points=control_points,
                weights=weights,
                spline=spline,
        )

        self._extractor = Extractor(self)
        self._proximity = Proximity(self)
        self._creator = Creator(self)

    @property
    def rationalbezier(self):
        """Returns same parametric representation as Rational Bezier Spline.

        Parameters
        ----------
        None

        Returns
        -------
        same : RationalBezier
        """
        return self.copy()

    @property
    def nurbs(self):
        """Returns same parametric representation as nurbs.

        Parameters
        -----------
        None

        Returns
        --------
        same_nurbs: NURBS
        """
        return NURBS(
                degrees=self.degrees,
                control_points=self.control_points,
                knot_vectors=[
                        [0] * (self.degrees[i] + 1) + [1]
                        * (self.degrees[i] + 1) for i in range(self.para_dim)
                ],
                weights=self.weights
        )


class BSpline(GustafSpline, splinepy.BSpline):

    def __init__(
            self,
            degrees=None,
            knot_vectors=None,
            control_points=None,
            spline=None,
    ):
        """BSpline of gustaf. Inherited from splinepy.BSpline and GustafSpline.

        Attributes
        -----------
        extract: Extractor
        create: Creator
        proximity: Proximity

        Parameters
        -----------
        degrees: (para_dim,) list-like
        knot_vectors: (para_dim, ...) list
        control_points: (m, dim) list-like

        Returns
        --------
        None
        """
        super(splinepy.BSpline, self).__init__(
                degrees=degrees,
                knot_vectors=knot_vectors,
                control_points=control_points,
                spline=spline,
        )

        self._extractor = Extractor(self)
        self._proximity = Proximity(self)
        self._creator = Creator(self)

    @property
    def bspline(self):
        """Returns same parametric representation as BSpline.

        Parameters
        -----------
        None

        Returns
        --------
        same_bspline : BSpline
        """
        return self.copy()

    @property
    def nurbs(self):
        """Returns same nurbs. Overwrites one from splinepy to return correct
        type.

        Parameters
        -----------
        None

        Returns
        --------
        same_nurbs: NURBS
        """
        from copy import deepcopy

        return NURBS(
                degrees=deepcopy(self.degrees),
                knot_vectors=deepcopy(self.knot_vectors),
                control_points=deepcopy(self.control_points),
                # fix dtype to match splinepy's dtype.
                weights=np.ones(self.control_points.shape[0], dtype="float64")
        )

    def extract_bezier_patches(self):
        """Overwrites splinepy-parent's function to ensure the conversion of
        Splines into a usable gustaf format.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        logging.warning(
                "Functionality not supported, please use:\n"
                "<BSpline>.extract.beziers()"
        )
        return None


class NURBS(GustafSpline, splinepy.NURBS):

    def __init__(
            self,
            degrees=None,
            knot_vectors=None,
            control_points=None,
            weights=None,
            spline=None,
    ):
        """NURBS of gustaf. Inherited from splinepy.NURBS.

        Attributes
        -----------
        extract: Extractor
        create: Creator
        proximity: Proximity

        Parameters
        -----------
        degrees: (para_dim,) list-like
        knot_vectors: (para_dim,) list
        control_points: (m, dim) list-like
        weights: (m, 1) list-like

        Returns
        --------
        None
        """
        super(splinepy.NURBS, self).__init__(
                degrees=degrees,
                knot_vectors=knot_vectors,
                control_points=control_points,
                weights=weights,
                spline=spline,
        )

        self._extractor = Extractor(self)
        self._proximity = Proximity(self)
        self._creator = Creator(self)

    @property
    def _mfem_ids(self):
        """Returns mfem index mapping. For ease of use.

        Parameters
        -----------
        None

        Returns
        --------
        None
        """
        if self.para_dim != 2:
            raise NotImplementedError(
                    "Sorry, only available for para_dim = 2 splines"
            )

        gustaf2mfem, mfem2gustaf = splinepy.io.mfem.mfem_index_mapping(
                self.para_dim,
                self.degrees,
                self.knot_vectors,
        )

        return gustaf2mfem, mfem2gustaf

    def extract_bezier_patches(self):
        """Overwrites splinepy-parent's function to ensure the conversion of
        Splines into a usable gustaf format.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        logging.warning(
                "Functionality not supported, please use:\n"
                "<NURBS>.extract.beziers()"
        )
        return None

    @property
    def nurbs(self):
        """Returns same parametric representation as nurbs.

        Parameters
        -----------
        None

        Returns
        --------
        same_nurbs: NURBS
        """
        return self.copy()


def from_mfem(nurbs_dict):
    """Construct a gustaf NURBS. Reorganizes control points and weights.

    Parameters
    -----------
    nurbs_dict: dict

    Returns
    --------
    nurbs: NURBS
    """
    _, m2gus = splinepy.io.mfem.mfem_index_mapping(
            len(nurbs_dict["degrees"]),
            nurbs_dict["degrees"],
            nurbs_dict["knot_vectors"],
    )

    return NURBS(
            degrees=nurbs_dict["degrees"],
            knot_vectors=nurbs_dict["knot_vectors"],
            control_points=nurbs_dict["control_points"][m2gus],
            weights=nurbs_dict["weights"][m2gus],
    )


def load_splines(fname):
    """Loads and creates gustaf NURBS. Does not perform any check or tests.

    Parameters
    -----------
    fname: str

    Returns
    --------
    gussplines: list
    """
    # first get dict_splines using splinepy
    dictsplines = splinepy.load_splines(fname, as_dict=True)

    # try to initialize with correct spline type
    gussplines = list()
    for dics in dictsplines:
        is_bspline = "knot_vectors" in dics
        is_nurbs = "weights" in dics

        if is_nurbs:
            gussplines.append(NURBS(**dics))
        elif is_bspline:
            gussplines.append(BSpline(**dics))
        else:
            gussplines.append(Bezier(**dics))

    return gussplines



from gustaf.create import vertices, faces, volumes

try:
    from gustaf.create import spline
except ImportError as err:
    # overwrites  all modules which depend on the `splinepy` library
    # with an object which will throw an error as soon
    # as it is used the first time. This means that any non spline based
    # functionality works as before, but as soon as these are used a
    # comprehensive exception will be raised which is understandable in
    # contrast to the possible multitude of errors previously possible
    from gustaf.helpers.raise_if import ModuleImportRaiser
    spline = ModuleImportRaiser("splinepy", err)

__all__ = [
        "vertices",
        "faces",
        "volumes",
        "spline",
]



"""gustaf/spline/__init__.py.

Interface for `splinepy` library and additional operations. Everything
related to `splinepy` is here, which includes its load function.
However, since create.spline does not rely on `splinepy`, it is not
here.
"""

from gustaf.spline import base
from gustaf.spline import create
from gustaf.spline import extract
from gustaf.spline.base import (
        Bezier, RationalBezier, BSpline, NURBS, show, from_mfem, load_splines
)
from gustaf.spline import ffd
from gustaf.spline import microstructure

import splinepy
from splinepy import io

# overwrite name to type map
splinepy.settings.NAME_TO_TYPE = dict(
        Bezier=Bezier,
        RationalBezier=RationalBezier,
        BSpline=BSpline,
        NURBS=NURBS,
)

# a shortcut
NAME_TO_TYPE = splinepy.settings.NAME_TO_TYPE

__all__ = [
        "base",
        "create",
        "extract",
        "Bezier",
        "RationalBezier",
        "BSpline",
        "NURBS",
        "show",
        "from_mfem",
        "load_splines",
        "ffd",
        "microstructure",
        "io",
]



from typing import Union

from gustaf.edges import Edges
from gustaf.faces import Faces
from gustaf.vertices import Vertices
from gustaf.volumes import Volumes

MESH_TYPES = Union[Vertices, Edges, Faces, Volumes]

try:
    from gustaf.spline.base import BSpline, NURBS, Bezier, RationalBezier
    SPLINE_TYPES = Union[Bezier, RationalBezier, BSpline, NURBS]
except ImportError:
    SPLINE_TYPES = None



"""gustaf/gustaf/ffd.py.

Freeform Deformation!


Adaptation of previous implementation in internal python package gustav by
Jaewook Lee.
"""
from typing import Any, List, Optional, Union
import numpy as np
from gustaf._base import GustafBase
from gustaf.show import show_vedo
from gustaf._typing import SPLINE_TYPES, MESH_TYPES
from gustaf.create.spline import with_bounds
from gustaf import settings


class FFD(GustafBase):

    def __init__(
            self,
            mesh: Optional[MESH_TYPES] = None,
            spline: Optional[SPLINE_TYPES] = None
    ):
        """
        Free-form deformation is a method used to deform an object by a
        deformation function. In our case the object is given via a mesh, the
        currently supported mesh-types are given by the variable
        :py:const:`gustaf._typing.MESH_TYPES`, and the deformations function
        by a spline, supported splines are given by the variable
        :py:const.:`gustaf._typing.SPLINE_TYPES`. The splines parametric
        dimension will be scaled in to a unit-hypercube as well as the
        original meshes vertices. The outline of the resulting mesh is given
        by the physical space of the spline.

        The FFD class provides functions to modify the spline by completely
        overwriting the spline whole spline or parts of it. To obtain the
        deformed mesh mapped into the latest spline, retrieve the mesh
        attribute.

        Please not that even though an object of the class can be initiated
        without a mesh, it is not possible to compute the deformation without
        one. Please ensure that at least a mesh is defined before retrieving
        the (deformed) mesh. If only a mesh is provided a default spline where
        the geometric dimensions have the bounds of the mesh is defined.

        A previously available partial FFD is currently not implemented, and
        is planned to be implemented in a separate class (LocalFFD).

        Parameters
        ----------
        mesh: Optional[MESH_TYPES]
            Mesh used in the FFD. Defaults to None.
        spline: Optional[SPLINE_TYPES]
            Spline used in the FFD. Defaults to None.

        Class Attributes
        ----------------
        _spline: SPLINE_TYPES
            Internal current spline
        _mesh: MESH_TYPES
            unscaled base mesh
        _q_vertices: np.ndarray (n, dim)
            Scaled vertices of the base mesh

        Returns
        -------
        None
        """
        # Use property definitions to store the values
        self._spline: SPLINE_TYPES = None
        self._mesh: MESH_TYPES = None
        self._o_mesh: MESH_TYPES = None
        self._q_vertices: np.ndarray = None

        if spline is not None:
            self.spline = spline
        if mesh is not None:
            self.mesh = mesh

        # self._is_calculated = False

    @property
    def mesh(self, ) -> MESH_TYPES:
        """Returns copy of current mesh. Before copying, it applies
        deformation.

        Returns
        -------
        MESH_TYPES
            Current Mesh with the deformation according to the current spline.
        """
        self._deform()
        return self._mesh.copy()

    @mesh.setter
    def mesh(self, mesh: MESH_TYPES):
        """Sets mesh. If it is first time, the copy of it will be saved as
        original mesh. If spline is already defined and in transformed status,
        it applies transformation directly.

        Parameters
        -----------
        mesh: MESH_TYPES
            Mesh used for the FFD

        Returns
        --------
        None
        """
        if self._spline is None:
            # Define a default spline if mesh is given but no spline
            par_dim = mesh.vertices.shape[1]
            self.spline = with_bounds(
                    [[0] * par_dim, [1] * par_dim], mesh.bounds()
            )

        self._logi("Setting mesh.")
        self._logi("Mesh Info:")
        self._logi("  Vertices: {v}.".format(v=mesh.vertices.shape))
        self._logi("  Bounds: {b}.".format(b=mesh.bounds()))
        self._o_mesh = mesh.copy()  # we keep original copy for visualization
        self._mesh = mesh.copy()  # another copy for current status.

        self._check_dimensions()

        self._scale_mesh_vertices()
        if self._spline:
            self._spline._data["gustaf_ffd_computed"] = False

    @property
    def spline(self):
        """Returns a copy of the spline. Please use the setter to explicitly
        make changes to the spline.

        Parameters
        -----------
        None

        Returns
        --------
        self._spline: Spline
        """
        return self._spline

    @spline.setter
    def spline(self, spline: SPLINE_TYPES):
        """Sets spline. The spline parametric range bounds will be converted
        into the bounds [0,1]^para_dim.

        Parameters
        -----------
        spline: SPLINE_TYPES
            New Spline for the next deformation

        Returns
        --------
        None
        """
        self._spline = spline

    def _check_dimensions(self) -> bool:
        """Checks if the dimension of the spline and the mesh match and

        Returns:
            bool: _description_
        """
        messages = []
        # Checks dimensions and ranges critical for a correct FFD calculation
        if self._spline and not self._spline.para_dim == self._spline.dim:
            messages.append(
                    "The parametric and geometric dimensions of the "
                    "spline are not the same."
            )
        if (
                self._spline and self._mesh
                and not self._spline.dim == self._mesh.vertices.shape[1]
        ):
            messages.append(
                    "The geometric dimensions of the spline and the "
                    "dimension of the mesh are not the same."
            )
        if len(messages) > 0:
            raise RuntimeError(
                    "Can not perform FFD due to spline and mesh "
                    "dimension mismatch. The following dimension mismatches:"
                    f"{messages}."
            )

    def _scale_mesh_vertices(self):
        """Scales the mesh vertices into the dimension of a hypercube and save
        them in self._q_vertices."""
        self._logd("Fitting mesh into spline's parametric space.")

        self._q_vertices = self._mesh.vertices.copy()

        original_mesh_bounds = self._mesh.bounds()

        # save mesh offset and scale for reasons
        self._mesh_offset = original_mesh_bounds[0]
        self._mesh_scale = 1 / (
                original_mesh_bounds[1] - original_mesh_bounds[0]
        )

        # scale and offset vertices coordinates
        self._q_vertices -= self._mesh_offset
        self._q_vertices *= self._mesh_scale

        self._logd("Successfully scaled and transformed mesh vertices!")

    def _deform(self):
        """Deforms mesh if spline or mesh changes were detected since last
        calculation. Meant for internal use.

        Parameters
        -----------
        None

        Returns
        --------
        None
        """
        if self._mesh is None or self._spline is None:
            raise RuntimeError(
                    "Can't perform deformation for the FFD, since either the "
                    "spline or(and) the mesh are not yet defined. "
                    "Please set either spline or mesh."
            )
        if self._spline._data.get("gustaf_ffd_computed", False):
            return None

        spline = self._spline.copy()
        if spline.has_knot_vectors:
            spline.normalize_knot_vectors()

        self._check_dimensions()

        self._logd("Applying FFD: Transforming vertices")

        # Here, we take _q_vertices, due to possible scale/offset.
        self._mesh.vertices = spline.evaluate(self._q_vertices)
        self._logd("FFD successful.")

        self._spline._data["gustaf_ffd_computed"] = True

    @property
    def control_points(self):
        """Returns current spline's control points. The control points can be
        directly updated with this.

        Returns
        --------
        self._spline.control_points: np.ndarray
        """
        return self._spline.control_points

    @control_points.setter
    def control_points(
            self, control_points: Union[List[List[float]], np.ndarray]
    ):
        """Sets control points and deforms mesh.

        Parameters
        -----------
        control_points: np.ndarray

        Returns
        --------
        None
        """
        assert self._spline.control_points.shape == control_points.shape,\
            "Given control points' shape does not match current ones!"

        self._spline.control_points = control_points.copy()
        self._logd("Set new control points.")

    def elevate_degree(self, *args, **kwargs):
        """Wrapper for Spline.elevate_degree.

        Parameters
        -----------
        *args:
        **kwargs:

        Returns
        --------
        None
        """
        self._spline.elevate_degree(*args, **kwargs)

    def insert_knots(self, parametric_dimension, knots):
        """Wrapper for Spline.insert_knots.

        Parameters
        -----------
        *args:
        **kwargs:

        Returns
        --------
        None
        """
        if "knot_vectors" in self._spline.required_properties:
            raise NotImplementedError(
                    "Can not perform knot insertion on Bezier spline."
            )
        self._spline.insert_knots(parametric_dimension, knots)

    def remove_knots(self, parametric_dimension, knots, tolerance=1e-8):
        """Wrapper for Spline.remove_knots.

        Parameters
        -----------
        *args:
        **kwargs:

        Returns
        --------
        None
        """
        if "knot_vectors" in self._spline.required_properties:
            raise NotImplementedError(
                    "Can not perform knot insertion on Bezier spline."
            )
        self._spline.remove_knots(
                parametric_dimension, knots, tolerance=tolerance
        )

    def reduce_degree(self, *args, **kwargs):
        """Wrapper for Spline.reduce_degree.

        Parameters
        -----------
        *args:
        **kwargs:

        Returns
        --------
        None
        """
        self._spline.reduce_degree(*args, **kwargs)

    def show(self, **kwargs) -> Any:
        """Visualize. Shows the deformed mesh and the current spline. Currently
        visualization is limited to vedo.

        Parameters
        ----------
        title: str
            Title of the vedo window. Defaults to "gustaf - FFD".
        return_showable: bool
            If true returns a dict of the showable items. Defaults to False.
        return_discrete: bool
            Return dict of gustaf discrete objects, for example,
            {Vertices, Edges, Faces}, instead of opening a window.
            Defaults to False.
        kwargs: Any
            Arbitrary keyword arguments. These are passed onto the vedo
            functions. Please be aware, that no checking of these are performed
            in this function.

        Returns
        -------
        Any:
            Returns, if applicable, the vedo plotter. 'close=False' as argument
            to get the plotter.
        """
        backend = kwargs.pop("backend", None)
        return_showable = kwargs.pop("return_showable", False)
        return_discrete = kwargs.pop("return_discrete", False)
        title = kwargs.pop("title", "gustaf - FFD")

        if return_discrete and return_showable:
            raise ValueError(
                    "Either one of following params can be True: "
                    "{return_discrete, return_showable} "
                    "You've set both True."
            )

        if backend is None:
            backend = settings.VISUALIZATION_BACKEND

        # prepare originals
        o_mesh = self._o_mesh.copy()
        # prepare deformed
        d_mesh = self.mesh  # copies

        things_to_show = dict()
        # let's show faces at most, since volumes can take awhile
        if o_mesh.kind == "volume":
            # only outer faces. overwrite
            o_mesh = o_mesh.tofaces(unique=False)
            o_mesh.update_faces(o_mesh.single_faces())
            d_mesh = d_mesh.tofaces(unique=False)
            d_mesh.update_faces(d_mesh.single_faces())

        # update meshes
        things_to_show.update(original_mesh=o_mesh)
        things_to_show.update(original_description="Original Mesh")
        things_to_show.update(deformed_mesh=d_mesh)
        things_to_show.update(deformed_description="Deformed Mesh with Spline")

        # update spline
        things_to_show.update(deformed_spline=self.spline)

        if return_discrete or not backend.startswith("vedo"):
            # spline is strictly not discrete.
            return things_to_show

        if return_showable:
            # let's turn everthing into showable and return
            for k, v in things_to_show.items():
                if isinstance(v, GustafBase):
                    things_to_show[k] = v.showable()

            return things_to_show

        # current workaround to set spline's surface alpha correctly
        # TODO: support this situation better
        spl = things_to_show.pop("deformed_spline")
        spl_showable = spl.showable(surface_alpha=.3)

        return show_vedo(
                [
                        things_to_show[k]
                        for k in things_to_show.keys()
                        if k.startswith("original")
                ],
                [
                        *[
                                things_to_show[k]
                                for k in things_to_show.keys()
                                if k.startswith("deformed")
                        ],
                        *spl_showable.values(),
                ],
                title=title,
        )

    def showable(self, **kwargs):
        """Returns a dictionary of showable items to describe the FFD at the
        current state.

        See show() for more information. This function redirects to it
        directly with the return_showable keyword set to True.
        """
        return self.show(return_showable=True, **kwargs)



"""gustaf/gustaf/settings.py.

Global variables/constants that's used throughout `gustaf`.
"""

TOLERANCE = 1e-10

FLOAT_DTYPE = "float64"
INT_DTYPE = "int32"

# OPTIONS are <"vedo" | "trimesh" | "matplotlib">
VISUALIZATION_BACKEND = "vedo"

VEDO_DEFAULT_OPTIONS = dict(
        vertex=dict(),
        edges=dict(),
        faces=dict(),
        volumes=dict(),
)

NTHREADS = 1



"""gustaf/spline/microstructure/tiles/__init__.py.

Interface for tools and generators creating simple microstructures.
"""

from gustaf.spline.microstructure.tiles import crosstile3d, \
    inversecrosstile3d, crosstile2d

from gustaf.spline.microstructure.tiles.crosstile3d import CrossTile3D
from gustaf.spline.microstructure.tiles.crosstile2d import CrossTile2D
from gustaf.spline.microstructure.tiles.inversecrosstile3d \
    import InverseCrossTile3D

__all__ = [
        "crosstile3d",
        "crosstile2d",
        "inversecrosstile3d",
        "CrossTile3D",
        "CrossTile2D",
        "InverseCrossTile3D",
]



from setuptools import setup

with open("gustaf/_version.py") as f:
    version = eval(f.read().strip().split("=")[-1])

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name="gustaf",
    version=version,
    description="Process and visualize geometries.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Jaewook Lee",
    author_email="jaewooklee042@gmail.com",
    url="https://github.com/tataratat/gustaf",
    packages=[
        "gustaf",
        "gustaf.utils",
        "gustaf.io",
        "gustaf.spline",
        "gustaf.create",
        "gustaf.helpers",
        "gustaf.spline.microstructure",
        "gustaf.spline.microstructure.tiles",
    ],
    install_requires=[
        "numpy",
    ],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering'
    ],
    license="MIT",
)



"""gustaf/gustaf/io/__init__.py.

io.
I - `load`.
O - `export`.
"""

from gustaf.io import ioutils
from gustaf.io import mfem
from gustaf.io import meshio
from gustaf.io import mixd
from gustaf.io import nutils
from gustaf.io.default import load

__all__ = [
        "ioutils",
        "mfem",
        "meshio",
        "mixd",
        "nutils",
        "load",
]



"""gustaf/gustaf/edges.py.

Edges. Also known as lines.
"""

import numpy as np

from gustaf import settings
from gustaf import utils
from gustaf import helpers
from gustaf.vertices import Vertices


class Edges(Vertices):

    kind = "edge"

    __slots__ = (
            "_edges",
            "_const_edges",
    )

    def __init__(
            self,
            vertices=None,
            edges=None,
            elements=None,
    ):
        """Edges. It has vertices and edges. Also known as lines.

        Parameters
        -----------
        vertices: (n, d) np.ndarray
        edges: (n, 2) np.ndarray
        """
        super().__init__(vertices=vertices)

        if edges is not None:
            self.edges = edges

        elif elements is not None:
            self.edges = elements

    @property
    def edges(self):
        """Returns edges. If edges is not its original property.

        Parameters
        -----------
        None

        Returns
        --------
        edges: (n, 2) np.ndarray
        """
        self._logd("returning edges")
        return self._edges

    @edges.setter
    def edges(self, es):
        """Edges setter. Similar to vertices, this is a tracked array.

        Parameters
        -----------
        es: (n, 2) np.ndarray

        Returns
        --------
        None
        """
        self._logd("setting edges")

        # shape check
        if es is not None:
            utils.arr.is_shape(es, (-1, 2), strict=True)

        self._edges = helpers.data.make_tracked_array(es, settings.INT_DTYPE)
        # same, but non-writeable view of tracked array
        self._const_edges = self._edges.view()
        self._const_edges.flags.writeable = False

    @property
    def const_edges(self):
        """Returns non-writeable version of edges.

        Parameters
        -----------
        None

        Returns
        --------
        const_edges (n, 2) np.ndarray
        """
        return self._const_edges

    @property
    def whatami(self):
        """whatami?

        Parameters
        -----------
        None

        Returns
        --------
        whatami: str
        """
        return "edges"

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def sorted_edges(self):
        """Sort edges along axis=1.

        Parameters
        -----------
        None

        Returns
        --------
        edges_sorted: (n_edges, 2) np.ndarray
        """
        edges = self._get_attr("edges")

        return np.sort(edges, axis=1)

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def unique_edges(self):
        """Returns a named tuple of unique edge info. Info includes unique
        values, ids of unique edges, inverse ids, count of each unique values.

        Parameters
        -----------
        None

        Returns
        --------
        unique_info: Unique2DIntegers
          valid attributes are {values, ids, inverse, counts}
        """
        unique_info = utils.connec.sorted_unique(
                self.sorted_edges(), sorted_=True
        )

        edges = self._get_attr("edges")

        # tuple is not assignable, but entry is mutable...
        unique_info.values[:] = edges[unique_info.ids]

        return unique_info

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def single_edges(self):
        """Returns indices of very unique edges: edges that appear only once.
        For well constructed faces, this can be considered as outlines.

        Parameters
        -----------
        None

        Returns
        --------
        outlines: (m,) np.ndarray
        """
        unique_info = self.unique_edges()

        return unique_info.ids[unique_info.counts == 1]

    @property
    def elements(self):
        """Returns current connectivity. A short cut in FE friendly term.
        Elements mean different things for different classes: Vertices ->
        vertices Edges -> edges Faces -> faces Volumes -> volumes.

        Parameters
        -----------
        None

        Returns
        --------
        elements: (n, d) np.ndarray
          int. iff elements=None
        """
        elem_name = type(self).__qualname__.lower()
        self._logd(f"returning {elem_name}")

        return getattr(self, elem_name)

    @elements.setter
    def elements(self, elems):
        """Calls corresponding connectivity setter. A short cut in FEM friendly
        term. Vertices -> vertices Edges -> edges Faces -> faces Volumes ->
        volumes.

        Parameters
        -----------
        elems: (n, d) np.ndarray

        Returns
        --------
        None
        """
        # naming rule in gustaf
        elem_name = type(self).__qualname__.lower()
        self._logd(f"seting {elem_name}'s connectivity.")
        return setattr(self, elem_name, elems)

    @property
    def const_elements(self):
        """Returns non-mutable version of elements.

        Parameters
        -----------
        None

        Returns
        --------
        non_mutable_elements: (n, d) TrackedArray
        """
        self._logd("returning const_elements")
        return getattr(self, "const_" + type(self).__qualname__.lower())

    @helpers.data.ComputedMeshData.depends_on(["vertices", "elements"])
    def centers(self):
        """Center of elements.

        Parameters
        -----------
        None

        Returns
        --------
        centers: (n_elements, d) np.ndarray
        """
        self._logd("computing centers")

        return self.const_vertices[self.const_elements].mean(axis=1)

    @helpers.data.ComputedMeshData.depends_on(["vertices", "elements"])
    def referenced_vertices(self, ):
        """Returns mask of referenced vertices.

        Parameters
        -----------
        None

        Returns
        --------
        referenced: (n,) np.ndarray
        """
        referenced = np.zeros(len(self.const_vertices), dtype=bool)
        referenced[self.const_elements] = True

        return referenced

    def remove_unreferenced_vertices(self):
        """Remove unreferenced vertices. Adapted from
        `github.com/mikedh/trimesh`

        Parameters
        -----------
        None

        Returns
        --------
        new_self: type(self)
        """
        referenced = self.referenced_vertices()

        inverse = np.zeros(len(self.vertices), dtype=settings.INT_DTYPE)
        inverse[referenced] = np.arange(referenced.sum())

        return self.update_vertices(
                mask=referenced,
                inverse=inverse,
        )

    def update_elements(self, mask):
        """Similar to update_vertices, but for elements.

        Parameters
        -----------
        mask: bool or (m,) np.ndarray

        Returns
        --------
        new_self: type(self)
        """
        self.elements = self.elements[mask]

        return self.remove_unreferenced_vertices()

    def update_edges(self, *args, **kwargs):
        """Alias to update_elements."""
        return self.update_elements(*args, **kwargs)

    def dashed(self, spacing=None):
        """Turn edges into dashed edges(=lines). Given spacing, it will try to
        chop edges as close to it as possible. Pattern should look:

        ``dashed edges``

        .. code-block::

             o--------o    o--------o    o--------o
             |<------>|             |<-->|
                (chop length)         (chop length / 2)

        Parameters
        -----------
        spacing: float
          Default is None and it will use self.bounds_diagonal_norm() / 50

        Returns
        --------
        dashing_edges: Edges
        """
        if self.kind != "edge":
            raise NotImplementedError("dashed is only for edges.")

        if spacing is None:
            # apply "automatic" spacing
            spacing = self.bounds_diagonal_norm() / 50

        v0s = self.vertices[self.edges[:, 0]]
        v1s = self.vertices[self.edges[:, 1]]

        distances = np.linalg.norm(v0s - v1s, axis=1)
        linspaces = (((distances // (spacing * 1.5)) + 1) * 3).astype(np.int32)

        # chop vertices!
        new_vs = []
        for v0, v1, lins in zip(v0s, v1s, linspaces):
            new_vs.append(np.linspace(v0, v1, lins))

        # we need all choped vertices.
        # there might be duplicating vertices. you can use merge_vertices
        new_vs = np.vstack(new_vs)
        # all mid points are explicitly defined, but they aren't required
        # so, rm.
        mask = np.ones(len(new_vs), dtype=bool)
        mask[1::3] = False
        new_vs = new_vs[mask]

        # prepare edges
        tmp_es = utils.connec.range_to_edges((0, len(new_vs)), closed=False)
        new_es = tmp_es[::2]

        return Edges(vertices=new_vs, edges=new_es)

    def shrink(self, ratio=.8, map_vertexdata=True):
        """Returns shrunk elements.

        Parameters
        -----------
        ratio: float
          Default is 0.8
        map_vertexdata: bool
          Default is True. Maps all vertexdata.

        Returns
        --------
        s_elements: Elements
          shrunk elements
        """
        elements = self.const_elements
        vs = np.vstack(self.vertices[elements])
        es = np.arange(len(vs))

        nodes_per_element = elements.shape[1]
        es = es.reshape(-1, nodes_per_element)

        mids = np.repeat(self.centers(), nodes_per_element, axis=0)

        vs -= mids
        vs *= ratio
        vs += mids

        s_elements = type(self)(vertices=vs, elements=es)

        if map_vertexdata:
            elements_flat = elements.ravel()
            for key, value in self.vertexdata.items():
                s_elements.vertexdata[key] = value[elements_flat]

            # probably wanna take visulation options too
            s_elements.vis_dict = self.vis_dict

        return s_elements

    def tovertices(self):
        """Returns Vertices obj.

        Parameters
        -----------
        None

        Returns
        --------
        vertices: Vertices
        """
        return Vertices(self.vertices)

    def _get_attr(self, attr):
        """Internal function to get attribute that maybe property or callable.
        Some properties are replaced by callable in subclasses as it may depend
        on other properties of subclass.

        Parameters
        -----------
        attr: str

        Returns
        --------
        attrib: Any
        """
        attrib = getattr(self, attr)
        return attrib() if callable(attrib) else attrib
